"""
This module provides a static factory class which maps the available services
for settings to the named constants in `SettingsServiceTypes`.

Authors:\n
- Philipp Schuette\n
"""

from pyzeal.pyzeal_types.settings_types import SettingsServicesTypes
from pyzeal.settings.json_settings_service import JSONSettingsService
from pyzeal.settings.settings_service import SettingsService


class SettingsServiceFactory:
    "Static factory class used to create instances of settings services."

    @staticmethod
    def getConcreteSettings(
        settingsType: SettingsServicesTypes = SettingsServicesTypes.DEFAULT,
    ) -> SettingsService:
        """
        Construct and return a new concrete implementation of the
        `SettingsService` interface to global PyZEAL settings.

        :param settingsType: type of settings service to construct
        """
        if settingsType == SettingsServicesTypes.JSON_SETTINGS:
            return JSONSettingsService()

        # return the current default settings provider
        return JSONSettingsService()



"""
This module contains tests of the grid-based Newton algorithm implementation
of the `FinderAlgorithm` interface.

Authors:\n
- Luca Wasmuth\n
"""

import numpy as np
import pytest

from pyzeal.algorithms.newton_grid import NewtonGridAlgorithm
from pyzeal.pyzeal_types.filter_types import FilterTypes
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.finder_test_cases import testFunctions
from pyzeal.tests.resources.utils import rootsMatchClosely
from pyzeal.utils.containers.rounding_container import RoundingContainer
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.root_context import RootContext
from pyzeal.utils.service_locator import ServiceLocator

settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)

# some test functions do not work due to algorithmic limitations
KNOWN_FAILURES = ["x^30", "x^50", "x^100", "1e6 * x^100"]


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("testName", sorted(testFunctions.keys()))
def testNewtonGridRootFinder(testName: str) -> None:
    """
    Test the Newton grid algorithm with the function given by `testName`

    :param testName: Name of the test case
    """
    if testName in KNOWN_FAILURES:
        pytest.skip()

    for numSamplePoints in [25, 35]:
        precision = testFunctions[testName].precision
        newtonGridAlgo = NewtonGridAlgorithm(numSamplePoints=numSamplePoints)
        container = RoundingContainer(precision=precision)
        ContainerFactory.registerPreDefinedFilter(
            container, filterType=FilterTypes.FUNCTION_VALUE_ZERO
        )
        ContainerFactory.registerPreDefinedFilter(
            container, filterType=FilterTypes.ZERO_IN_BOUNDS
        )
        context = RootContext(
            f=testFunctions[testName].testFunc,
            df=testFunctions[testName].testFuncDerivative,
            container=container,
            precision=precision,
            reRan=testFunctions[testName].reRan,
            imRan=testFunctions[testName].imRan,
        )
        newtonGridAlgo.calcRoots(context)
        foundRoots = container.getRoots()
        expectedRoots = np.array(testFunctions[testName].expectedRoots)

        assert rootsMatchClosely(
            foundRoots, expectedRoots, precision=precision
        )



"""
The `settings` framework of `PyZEAL`.
"""

from pyzeal.settings.invalid_setting_exception import InvalidSettingException
from pyzeal.settings.settings_service import SettingsService

__all__ = [
    "InvalidSettingException",
    "SettingsService",
]



"""
Test the JSONHelper for correct exception throwing behavior.
"""
from unittest.mock import MagicMock, patch

import pytest

from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.invalid_setting_exception import InvalidSettingException
from pyzeal.settings.json_helper import (
    JSONHelper,
    tCoreSettingsKey,
    tSettingsKey,
)


@patch("pyzeal.settings.json_helper.dump")
@patch("builtins.open")
@patch("pyzeal.settings.json_helper.load")
@pytest.mark.parametrize(
    "settingName",
    [
        "INVALID_KEY",
        "defaultAlgorithm",
        "defaultContainer",
        "defaultEstimator",
    ],
)
def testInvalidCoreSettings(
    loadMock: MagicMock,
    openMock: MagicMock,
    dumpMock: MagicMock,
    settingName: tCoreSettingsKey,
) -> None:
    """
    Test exception raising when an invalid core setting is given.

    :param settingName: Setting name to test, parametrized by pytest
    """
    with pytest.raises(InvalidSettingException):
        # at least one of these should fail, no matter which
        # settingName is used
        JSONHelper.createOrUpdateCoreSetting(
            "DOES_NOT_EXIST.json",
            settingName,
            AlgorithmTypes.NEWTON_GRID,
        )
        JSONHelper.createOrUpdateCoreSetting(
            "DOES_NOT_EXIST.json",
            settingName,
            EstimatorTypes.QUADRATURE_ESTIMATOR,
        )
    if settingName != "defaultAlgorithm":
        dumpMock.assert_not_called()
    else:
        dumpMock.assert_called_once()
    openMock.assert_called()
    loadMock.assert_called()


@patch("pyzeal.settings.json_helper.dump")
@patch("builtins.open")
@patch("pyzeal.settings.json_helper.load")
@pytest.mark.parametrize(
    "settingName", ["INVALID_KEY", "logLevel", "verbose", "precision"]
)
def testInvalidSettings(
    loadMock: MagicMock,
    openMock: MagicMock,
    dumpMock: MagicMock,
    settingName: tSettingsKey,
) -> None:
    """
    Test exception raising when an invalid setting is given.

    :param settingName: Setting name to test, parametrized by pytest
    """
    with pytest.raises(InvalidSettingException):
        # at least one of these should fail, no matter which
        # settingName is used
        JSONHelper.createOrUpdateSetting(
            "DOES_NOT_EXIST.json", settingName, False
        )
        JSONHelper.createOrUpdateSetting(
            "DOES_NOT_EXIST.json", settingName, (1, 1)
        )
    if settingName != "verbose":
        dumpMock.assert_not_called()
    else:
        dumpMock.assert_called_once()
    openMock.assert_called()
    loadMock.assert_called()



"""
Class `CoreSettingsService` from the package `pyzeal_settings`.
This module defines a protocol for a generic settings provider pertaining to
the core features of `PyZEAL`.

Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol, runtime_checkable

from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes


@runtime_checkable
class CoreSettingsService(Protocol):
    """
    Class providing a layer of abstraction for storage and retrieval of PyZEAL
    core settings. Concrete `SettingService` implementations can choose freely
    their data model and persistence layer.
    """

    @property
    def defaultContainer(self) -> ContainerTypes:
        """
        Get the default container type.

        :return: Default container type
        """
        ...

    @defaultContainer.setter
    def defaultContainer(self, value: ContainerTypes) -> None:
        """
        Set the default container type to `value`.

        :param value: New default container type.
        """
        ...

    @property
    def defaultAlgorithm(self) -> AlgorithmTypes:
        """
        Get the default algorithm type.

        :return: Default algorithm type.
        """
        ...

    @defaultAlgorithm.setter
    def defaultAlgorithm(self, value: AlgorithmTypes) -> None:
        """
        Set the default algorithm type to `value`.

        :param value: New default algorithm type.
        """
        ...

    @property
    def defaultEstimator(self) -> EstimatorTypes:
        """
        Get the default estimator type.

        :return: Default estimator type.
        """
        ...

    @defaultEstimator.setter
    def defaultEstimator(self, value: EstimatorTypes) -> None:
        """
        Set the default estimator type.

        :param value: New default estimator type.
        """
        ...



"""
This module tests the behavior of the quadratue-based estimator.
"""
import numpy as np
import pytest

from pyzeal.algorithms.estimators import EstimatorCache
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.tests.resources.estimator_resources import (
    lineCases,
    rectangleCases,
)
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.factories.estimator_factory import EstimatorFactory
from pyzeal.utils.root_context import RootContext


@pytest.mark.parametrize("testName", sorted(rectangleCases.keys()))
def testQuadratueEstimatorRectangle(testName: str) -> None:
    """
    Test the quadrature-based estimator over a rectangular contour with the
    test case given by `testName`.

    :param testName: Case to test.
    """
    context, order, expected = rectangleCases[testName]
    est = EstimatorFactory.getConcreteEstimator(
        EstimatorTypes.QUADRATURE_ESTIMATOR,
        numPts=6500,
        deltaPhi=0.01,
        maxPrecision=1e-10,
        cache=EstimatorCache(),
    )
    result = est.calcMoment(order, context.reRan, context.imRan, context)
    assert np.abs(result - expected) < 1e-6


@pytest.mark.parametrize("testName", sorted(lineCases.keys()))
def testQuadratureEstimatorLine(testName: str) -> None:
    """
    Test the quadrature-based estimator over a line with the test case given
    by `testName`.

    :param testName: Case to test.
    """
    context, order, zStart, zEnd, expected = lineCases[testName]
    est = EstimatorFactory.getConcreteEstimator(
        EstimatorTypes.QUADRATURE_ESTIMATOR,
        numPts=6500,
        deltaPhi=0.01,
        maxPrecision=1e-10,
        cache=EstimatorCache(),
    )
    result = est.calcMomentAlongLine(order, zStart, zEnd, context)
    assert np.abs(result - expected) < 1e-6


def testExceptionDerivativefree() -> None:
    """
    Test exception throwing if the quadrature estimator is not provided with
    the derivative.
    """
    context = RootContext(
        f=lambda x: x,
        df=None,
        container=ContainerFactory.getConcreteContainer(),
        precision=(3, 3),
    )
    est = EstimatorFactory.getConcreteEstimator(
        EstimatorTypes.QUADRATURE_ESTIMATOR,
        numPts=6500,
        deltaPhi=0.01,
        maxPrecision=1e-10,
        cache=EstimatorCache(),
    )
    with pytest.raises(ValueError):
        est.calcMomentAlongLine(0, 0, 1, context)



"""
Class PyZEALLogger from the package pyzeal_logging.

This module provides a facade for `logging.Logger` class from the standard
library `logging` module. Any access to loggers in this project should happen
only through this interface.

Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol, Union


class PyZEALLogger(Protocol):
    "The logging interface used throughout the PyZEAL project."

    def debug(self, msg: str, *args: Union[str, int, float]) -> None:
        """
        Log `msg` with severity DEBUG.

        :param msg: Messageto log
        """
        ...

    def info(self, msg: str, *args: Union[str, int, float]) -> None:
        """
        Log `msg` with severity INFO.

        :param msg: Messageto log
        """
        ...

    def warning(self, msg: str, *args: Union[str, int, float]) -> None:
        """
        Log `msg` with severity WARNING.

        :param msg: Messageto log
        """
        ...

    def error(self, msg: str, *args: Union[str, int, float]) -> None:
        """
        Log `msg` with severity ERROR.

        :param msg: Messageto log
        """
        ...

    def critical(self, msg: str, *args: Union[str, int, float]) -> None:
        """
        Log `msg` with severity CRITICAL.

        :param msg: Messageto log
        """
        ...

    def setLevel(self, level: int) -> None:
        """
        Set the logging level to `level`

        :param level: `LogLevel` to set
        """
        ...

    def isEnabledFor(self, level: int) -> bool:
        """
        Test if logging is enabled for level `level`.

        :param level: Level to test
        :return: If logging is enabled for level `level`
        """
        ...



"""
Class AssociatedPolynomialAlgorithm from the package pyzeal_algorithms.

This module defines a root finding algorithm based on associated polynomials
whose coefficients are calculated from higher moments of the logarithmic
derivative using Newton's identities. Our implementation follows the original
ideas of [Delves, Lyness].

Authors:\n
- Philipp Schuette\n
"""

from typing import Tuple

import numpy as np

from pyzeal.algorithms.constants import FOUR_PI, TWO_PI
from pyzeal.algorithms.simple_holo import SimpleArgumentAlgorithm
from pyzeal.utils.root_context import RootContext


class AssociatedPolynomialAlgorithm(SimpleArgumentAlgorithm):
    """
    Class representation of a root finding algorithm which uses numerical
    quadrature to calculate higher moments. Then Newton's identities can be
    used to calculate the associated polynomial from these moments. The zeros
    of the latter coincide with the zeros of the original target function.
    """

    def decideRefinement(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        phi: float,
        context: RootContext,
    ) -> None:
        """
        Decide which way the current search area should be subdivided and
        calculate roots in the subdivided areas. The simple strategy consists
        of the choices (1) return, if argument indicates no roots, (2) place
        root in `context.container` if roots present and accuracy attained,
        (3) construct and solve Newton identities if only a few roots present,
        or (4) subdivide further if too many roots present.

        :param reRan: Real part of current search Range
        :param imRan: Imaginary part of current search range
        :param phi: Change in argument along the boundary of the current range
        :param context: `RootContext` in which the algorithm operates
        """
        if context.df is None:
            raise ValueError(
                "holomorphic Newton polynomial algorithm needs the derivative!"
            )
        # calculate difference between right/left and top/bottom
        x1, x2 = reRan
        y1, y2 = imRan
        deltaRe = x2 - x1
        deltaIm = y2 - y1

        # check if desired accuracy is aquired
        epsReal = 10 ** (-context.precision[0])
        epsImag = 10 ** (-context.precision[1])
        if phi > TWO_PI and deltaRe < epsReal and deltaIm < epsImag:
            SimpleArgumentAlgorithm.getRootFromRectangle(
                x2, x1, y2, y1, phi, context
            )
            return

        # check if the current box contains sufficiently few roots to construct
        # an associated polynomial with stable coefficients/roots
        if TWO_PI < phi < FOUR_PI:
            self.logger.debug(
                "constructing Newton polynomials from higher moments!"
            )

            firstMoment = self.estimator.calcMoment(
                order=1, reRan=(x1, x2), imRan=(y1, y2), context=context
            )
            newRoot = firstMoment / (2 * np.pi)
            context.container.addRoot(
                (newRoot, int(np.round(phi / (2 * np.pi)))),
                context.toFilterContext(),
            )

            if context.progress is not None and context.task is not None:
                context.progress.update(
                    context.task, advance=deltaRe * deltaIm
                )
            return

        super().decideRefinement((x1, x2), (y1, y2), phi, context)



"""
This module provides a static factory class which maps the algorithms available
in the `AlgorithmTypes` enumeration to appropriate members of the
`FinderAlgorithm` type.

Authors:\n
- Philipp Schuette\n
"""

from typing import Optional

from pyzeal.algorithms.finder_algorithm import FinderAlgorithm
from pyzeal.algorithms.newton_grid import NewtonGridAlgorithm
from pyzeal.algorithms.polynomial_holo import AssociatedPolynomialAlgorithm
from pyzeal.algorithms.simple_holo import SimpleArgumentAlgorithm
from pyzeal.algorithms.simple_holo_newton import SimpleArgumentNewtonAlgorithm
from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_logging.log_manager import LogManager
from pyzeal.pyzeal_logging.logger_facade import PyZEALLogger
from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.service_locator import ServiceLocator


class AlgorithmFactory:
    "Static factory class used to create instances of root finding algorithms."

    # the module level logger
    _logger: Optional[PyZEALLogger] = None

    @staticmethod
    def getConcreteAlgorithm(
        algoType: AlgorithmTypes = AlgorithmTypes.DEFAULT,
        *,
        estimatorType: EstimatorTypes = EstimatorTypes.DEFAULT,
        numSamplePoints: Optional[int] = None,
    ) -> FinderAlgorithm:
        """
        Construct and return an algorithm instance based on the given type of
        algorithm `algoType`. Additional, algorithm specific configuration
        arguments are optional.

        :param algoType: type of algorithm to construct
        :param estimatorType: type of argument estimator to use
        :param numSamplePoints: sample point configuration for NewtonGridAlgo
        :return: a concrete `FinderAlgorithm` instance
        """
        if AlgorithmFactory._logger is None:
            AlgorithmFactory._logger = LogManager.initLogger(
                __name__.rsplit(".", maxsplit=1)[-1],
                ServiceLocator.tryResolve(SettingsService).logLevel,
            )
        if algoType == AlgorithmTypes.NEWTON_GRID:
            AlgorithmFactory._logger.debug(
                "requested usage of a NewtonGridAlgorithm..."
            )
            if numSamplePoints:
                return NewtonGridAlgorithm(numSamplePoints=numSamplePoints)
            return NewtonGridAlgorithm()
        if algoType == AlgorithmTypes.SIMPLE_ARGUMENT:
            AlgorithmFactory._logger.debug(
                "requested usage of a SimpleArgumentAlgorithm..."
            )
            return SimpleArgumentAlgorithm(estimatorType=estimatorType)
        if algoType == AlgorithmTypes.SIMPLE_ARGUMENT_NEWTON:
            AlgorithmFactory._logger.debug(
                "requested usage of a SimpleArgumentNewtonAlgorithm..."
            )
            return SimpleArgumentNewtonAlgorithm(estimatorType=estimatorType)
        if algoType == AlgorithmTypes.ASSOCIATED_POLYNOMIAL:
            AlgorithmFactory._logger.debug(
                "requested usage of an AssociatedPolynomialAlgorithm..."
            )
            return AssociatedPolynomialAlgorithm(estimatorType=estimatorType)

        # return the current default algorithm
        AlgorithmFactory._logger.debug(
            "requested usage of the default algorithm..."
        )
        settings = ServiceLocator.tryResolve(SettingsService)
        return AlgorithmFactory.getConcreteAlgorithm(
            settings.defaultAlgorithm,
            numSamplePoints=numSamplePoints,
            estimatorType=estimatorType,
        )

    @staticmethod
    def setLevel(level: LogLevel) -> None:
        """
        Set the log level.

        :param level: the new log level
        """
        if AlgorithmFactory._logger is not None:
            AlgorithmFactory._logger.setLevel(level=level.value)



"""
This module provides named constants used to identify a predefined set of
filters to be applied with root containers. These filters should be applied
via the `ContainerFactory` static class and removed via the respective method
on container or root finder instances.

Authors:\n
- Philipp Schuette\n
"""

from enum import Enum


class FilterTypes(Enum):
    "Enumeration containing named constants identifying available filters."
    FUNCTION_VALUE_ZERO = "[function-value-zero]"
    ZERO_IN_BOUNDS = "[zero-in-bounds]"



"""
The `cli` interface to the `PyZEAL` package.

Right now the command line interface exposes functionality related to

- viewing and changing of options
- viewing and (un-)installing of plugins
- viewing basic information like help and package version
"""



"""
This module provides a simple argument estimator based on summation of
incremental changes in the phase of the argument.

Authors:\n
- Philipp Schuette\n
"""

from typing import Dict, Literal, Tuple, Union, cast

import numpy as np

from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.estimators.constants import MAX_Z_LENGTH, Z_REFINE
from pyzeal.algorithms.estimators.estimator_cache import EstimatorCache
from pyzeal.pyzeal_types.root_types import tVec
from pyzeal.utils.root_context import RootContext

# type alias for internal caches
tCache = Dict[
    Tuple[float, int],
    Dict[
        Tuple[float, Union[Literal["start"], Literal["end"]]],
        Tuple[tVec, tVec],
    ],
]


class SummationEstimator(ArgumentEstimator):
    """
    Class implementation of a simple argument estimator.

    It is based on discretizing the path of integration and summing the change
    of argument between these points to compute the integral of the logarithmic
    derivative.
    """

    __slots__ = (
        "numPts",
        "deltaPhi",
        "maxPrecision",
        "_cache",
        "cacheHorizontal",
        "cacheVertical",
    )

    def __init__(
        self,
        *,
        numPts: int,
        deltaPhi: float,
        maxPrecision: float,
        cache: EstimatorCache,
    ) -> None:
        """
        Initialize a `SummationEstimator` with given settings.

        :param numPts: Number of points used for estimation.
        :param deltaPhi: Threshold for argument change. When the change in
            argument between two points exceeds this, it will be further
            refined.
        :param maxPrecision: Maximum precision for refinement
        :param cache: Cache to store intermediate computation results.
        """
        self.numPts = numPts
        self.deltaPhi = deltaPhi
        self.maxPrecision = maxPrecision
        self._cache = cache
        # initialize additional internal caches to avoid function re-evals
        self.cacheHorizontal: tCache = {}
        self.cacheVertical: tCache = {}

        self.logger.info("initialized new phase summation based estimator...")

    # docstr-coverage:inherited
    @property
    def cache(self) -> EstimatorCache:
        return self._cache

    def calcMomentAlongLine(
        self,
        order: int,
        zStart: complex,
        zEnd: complex,
        context: RootContext,
    ) -> complex:
        r"""
        The result of this method coincides exactly with the integral of the
        logarithmic derivative as long as the increments remain below the
        threshold of :math:`\pi`. It can be proven that local function
        information is not sufficient to verify this condition in general.

        :param order: Moment to calculate
        :param zStart: Starting point of the line
        :param zEnd: End point of the line
        :param context: `RootContext` containing the necessary information
        :raises ValueError: An error is raised if the line is not
            parallel to either the real or imaginary axis.
        :return: The moment as calculated along the given line.
        """
        # look for required function values in the internal caches
        x1, y1 = zStart.real, zStart.imag
        x2, y2 = zEnd.real, zEnd.imag
        phi: complex
        # handle the case of horizontal line first
        if y1 == y2:
            phi = self.retrieveCachedHorizontal(order, x1, x2, y1, context)
        elif x1 == x2:
            phi = self.retrieveCachedVertical(order, y1, y2, x1, context)
        else:
            raise ValueError(
                f"{zStart} and {zEnd} must define an axis-parallel line!"
            )

        return phi

    def retrieveCachedHorizontal(
        self, order: int, x1: float, x2: float, y: float, context: RootContext
    ) -> complex:
        """
        Try to retrieve the moment along a horizontal line from cache,
        calculating it if necessary.

        :param order: Moment to retrieve
        :param x1: Starting point x-value
        :param x2: End point x-value
        :param y: y-value of the line
        :param context: `RootContext` containing the necessary information
        :return: The moment as calculated along the given line.
        """
        cache = self.cacheHorizontal
        # the internal caches only contain positively oriented lines
        sign = 1 if x1 < x2 else -1
        x1, x2 = sorted((x1, x2))
        if (y, order) in cache:
            if (x1, "start") in cache[(y, order)]:
                self.logger.debug(
                    "horizontal line start in internal cache found - dividing!"
                )
                value = cache[(y, order)][(x1, "start")]
                if len((indices := np.where(np.real(value[0]) <= x2)[0])) < 3:
                    self.logger.warning("must regenerate support points!")
                    newValue = self.genPhiArr(
                        order, x1 + 1j * y, x2 + 1j * y, context
                    )
                else:
                    middleIdx = indices[-1]
                    newValue = (
                        value[0][: middleIdx + 1],
                        value[1][: middleIdx + 1],
                    )
                self.storeCache(y, order, x1, x2, cache, newValue)
                return cast(float, sign * newValue[1].sum())
            if (x2, "end") in cache[(y, order)]:
                self.logger.debug(
                    "horizontal line end in internal cache found - dividing!"
                )
                value = cache[(y, order)][(x2, "end")]

                if len((indices := np.where(x1 <= np.real(value[0]))[0])) < 3:
                    self.logger.warning("must regenerate support points!")
                    newValue = self.genPhiArr(
                        order, x1 + 1j * y, x2 + 1j * y, context
                    )
                else:
                    middleIdx = indices[0]
                    newValue = (
                        value[0][middleIdx:],
                        value[1][middleIdx:],
                    )
                self.storeCache(y, order, x1, x2, cache, newValue)
                return cast(float, sign * newValue[1].sum())

        self.logger.debug("internal cache miss on horizontal line!")
        newValue = self.genPhiArr(order, x1 + 1j * y, x2 + 1j * y, context)

        if (y, order) not in cache:
            cache[(y, order)] = {}
        self.storeCache(y, order, x1, x2, cache, newValue)
        return cast(float, sign * newValue[1].sum())

    def retrieveCachedVertical(
        self, order: int, y1: float, y2: float, x: float, context: RootContext
    ) -> complex:
        """
        Try to retrieve the moment along a vertical line from cache,
        calculating it if necessary.

        :param order: Moment to retrieve
        :param y1: Starting point y-value
        :param y2: End point y-value
        :param x: x-value of the line
        :param context: `RootContext` containing the necessary information.
        :return: The moment as calculated along the given line.
        """
        cache = self.cacheVertical
        sign = 1 if y1 < y2 else -1
        y1, y2 = sorted((y1, y2))
        if (x, order) in cache:
            if (y1, "start") in cache[(x, order)]:
                self.logger.debug(
                    "vertical line start in internal cache found - dividing!"
                )
                value = cache[(x, order)][(y1, "start")]
                if len((indices := np.where(np.imag(value[0]) <= y2)[0])) < 3:
                    self.logger.warning("must regenerate support points!")
                    newValue = self.genPhiArr(
                        order, x + 1j * y1, x + 1j * y2, context
                    )
                else:
                    middleIdx = indices[-1]
                    newValue = (
                        value[0][: middleIdx + 1],
                        value[1][: middleIdx + 1],
                    )
                self.storeCache(x, order, y1, y2, cache, newValue)
                return cast(float, sign * newValue[1].sum())
            if (y2, "end") in cache[(x, order)]:
                self.logger.debug(
                    "vertical line end in internal cache found - dividing!"
                )
                value = cache[(x, order)][(y2, "end")]
                if len((indices := np.where(np.imag(value[0]) <= y2)[0])) < 3:
                    self.logger.warning("must regenerate support points!")
                    newValue = self.genPhiArr(
                        order, x + 1j * y1, x + 1j * y2, context
                    )
                else:
                    middleIdx = np.where(y1 <= np.imag(value[0]))[0][0]
                    newValue = (
                        value[0][middleIdx:],
                        value[1][middleIdx:],
                    )
                self.storeCache(x, order, y1, y2, cache, newValue)
                return cast(float, sign * newValue[1].sum())

        self.logger.debug("internal cache miss on vertical line!")
        newValue = self.genPhiArr(order, x + 1j * y1, x + 1j * y2, context)

        if (x, order) not in cache:
            cache[(x, order)] = {}
        self.storeCache(x, order, y1, y2, cache, newValue)
        return cast(float, sign * newValue[1].sum())

    def genPhiArr(
        self,
        order: int,
        zStart: complex,
        zEnd: complex,
        context: RootContext,
    ) -> Tuple[tVec, tVec]:
        """
        Calculate an array of complex argument values from the function values
        of the target function `context.f` on the complex line
        `[zStart, zEnd]`. Zeros of the target function found during this
        procedure are put into `context.container` immediately and the complex
        line is adjusted by translating into direction `pos` by a small offset.
        The number of support points on the line is adjusted dynamically.

        :param order: The moment which is to be calculated.
        :param zStart: Starting point of the line segment.
        :param zEnd: End point of the line segment.
        :param context: Context of the current calculation.
        :return: Points along with estimated change of argument between them.
        """
        if order != 0:
            raise NotImplementedError(
                f"summation estimator is not implemented for order={order}>0!"
            )

        # build the array f(z_{k+1})/f(z_k) of quotients of successive values
        zArr, funcArr = self.genFuncArr(zStart, zEnd, context, self.numPts)
        funcArr = funcArr[1:] / funcArr[:-1]
        # compute change in argument between two points on the line
        phiArr = np.arctan2(funcArr.imag, funcArr.real)

        # loop over entries in phiArr larger than deltaPhi with dynamically
        # increasing refinement size
        idxPhi = np.where(abs(phiArr) >= self.deltaPhi)[0]
        idx = 0
        while idxPhi.size > 0:
            idx += 1
            # refine grid between z values at large phi values
            k = idxPhi[0]
            self.logger.debug(
                "Refining the line between [%s, %s] because phase change=%s",
                str(zArr[k]),
                str(zArr[k + 1]),
                str(phiArr[k]),
            )
            if abs(zArr[k] - zArr[k + 1]) < self.maxPrecision:
                self.logger.warning("maximum z-refinement depth reached!")
                break
            if len(zArr) > MAX_Z_LENGTH * self.numPts:
                self.logger.warning("maximum z-length reached!")
                break
            refinementFactor = int(
                idx * Z_REFINE * abs(phiArr[k]) / self.deltaPhi
            )
            zArrRefinement, funcArr = self.genFuncArr(
                cast(complex, zArr[k]),
                cast(complex, zArr[k + 1]),
                context,
                refinementFactor,
            )
            funcArr = funcArr[1:] / funcArr[:-1]
            phiArrRefinement = np.arctan2(funcArr.imag, funcArr.real)

            # concatenate old and new arrays
            zArr = np.concatenate(
                (zArr[: k + 1], zArrRefinement[1:-1], zArr[k + 1 :])
            )

            phiArr = np.concatenate(
                (phiArr[:k], phiArrRefinement, phiArr[k + 1 :])
            )
            idxPhi = np.where(abs(phiArr) >= self.deltaPhi)[0]

        return zArr, phiArr

    def storeCache(
        self,
        z: float,
        order: int,
        start: float,
        end: float,
        cache: tCache,
        newValue: Tuple[tVec, tVec],
    ) -> None:
        """
        Store a new value in cache.

        :param z: Real or imaginary part of the line, depending on if the line
            is horizontal or vertical.
        :param order: Order of the moment to be calculated
        :param start: Real or imaginary part of the starting point, depending
            on line orientation
        :param end: Real or imaginary part of the starting point, depending
            on line orientation
        :param cache: Cache to store the new value in.
        :param newValue: New value to store.
        """
        cache[(z, order)][(start, "start")] = newValue
        cache[(z, order)][(end, "end")] = newValue



"""
A simple in-memory implementation of the `SettingsService` interface. Instances
of this service do not persist settings beyond program runtime. Their main use
is during testing of components which depend on settings.

Authors:\n
- Philipp Schuette\n
"""

from typing import Tuple

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.settings_service import SettingsService


class RAMSettingsService(SettingsService):
    "Simple, non-persistent implementation of `SettingsService`."

    def __init__(
        self,
        container: ContainerTypes = ContainerTypes.ROUNDING_CONTAINER,
        algorithm: AlgorithmTypes = AlgorithmTypes.SIMPLE_ARGUMENT,
        estimator: EstimatorTypes = EstimatorTypes.SUMMATION_ESTIMATOR,
        logLevel: LogLevel = LogLevel.NOTSET,
        precision: Tuple[int, int] = (0, 0),
        verbose: bool = False,
    ) -> None:
        """
        Initialize a new in-memory settings service with given default values.

        :param :
        :param :
        :param :
        :param :
        :param :
        :param :
        """
        self._container = container
        self._algorithm = algorithm
        self._estimator = estimator
        self._logLevel = logLevel
        self._precision = precision
        self._verbose = verbose

    # docstr-coverage:inherited
    @property
    def defaultContainer(self) -> ContainerTypes:
        return self._container

    # docstr-coverage:inherited
    @defaultContainer.setter
    def defaultContainer(self, value: ContainerTypes) -> None:
        self._container = value

    # docstr-coverage:inherited
    @property
    def defaultAlgorithm(self) -> AlgorithmTypes:
        return self._algorithm

    # docstr-coverage:inherited
    @defaultAlgorithm.setter
    def defaultAlgorithm(self, value: AlgorithmTypes) -> None:
        self._algorithm = value

    # docstr-coverage:inherited
    @property
    def defaultEstimator(self) -> EstimatorTypes:
        return self._estimator

    # docstr-coverage:inherited
    @defaultEstimator.setter
    def defaultEstimator(self, value: EstimatorTypes) -> None:
        self._estimator = value

    # docstr-coverage:inherited
    @property
    def logLevel(self) -> LogLevel:
        return self._logLevel

    # docstr-coverage:inherited
    @logLevel.setter
    def logLevel(self, value: LogLevel) -> None:
        self._logLevel = value

    # docstr-coverage:inherited
    @property
    def verbose(self) -> bool:
        return self._verbose

    # docstr-coverage:inherited
    @verbose.setter
    def verbose(self, value: bool) -> None:
        self._verbose = value

    # docstr-coverage:inherited
    @property
    def precision(self) -> Tuple[int, int]:
        return self._precision

    # docstr-coverage:inherited
    @precision.setter
    def precision(self, value: Tuple[int, int]) -> None:
        self._precision = value



"""
Module `service_configuration_exception.py` from the package `PyZEAL`.

This module contains the generic exception signaling that an invalid
configuration has been received.

Authors:\n
- Philipp Schuette\n
"""

from typing import Generic, Type, TypeVar

T = TypeVar("T")


class InvalidServiceConfiguration(Exception, Generic[T]):
    """
    This exception should be raised if invalid configurations are encountered
    during registration/resolving of services with ServiceLocator.
    """

    def __init__(self, serviceType: Type[T]) -> None:
        """
        Initialize a new `InvalidServiceConfiguration` exception.

        :param serviceType: the faulty service
        """
        super().__init__("invalid service request: " + str(serviceType))



"""
Timing benchmark suite for the newton grid based root finding algorithm.

Authors:\n
- Luca Wasmuth\n
"""

from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.finder_helpers import newtonGridFinder
from pyzeal.tests.resources.finder_test_cases import testFunctions
from pyzeal.utils.service_locator import ServiceLocator

benchmarkFunctions = [
    "x^2-1",
    "x^2+1",
    "x^4-1",
    "x^5-4x+2",
    "sin(x)",
    "exp(x)",
]

# disable progress bar
settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)


class NewtonGridSuite:
    """
    Timing benchmarks for the root finding algorithm which uses a grid of
    starting points for an ordinary Newton algorithm. Includes simple
    polynomials as well as trigonometric and exponential functions.
    """

    def TimeNewtonGridRootfinder(self) -> None:
        """
        Runs the Newton grid based version of the rootfinding algorithm for all
        test cases and non-parallel.
        """
        for caseName in benchmarkFunctions:
            gridRF = newtonGridFinder(
                caseName, numSamplePoints=50, parallel=False
            )
            reRan = testFunctions[caseName].reRan
            imRan = testFunctions[caseName].imRan
            gridRF.calculateRoots(reRan, imRan, precision=(5, 5))

    def TimeNewtonGridRootfinderDerivativeFree(self) -> None:
        """
        Runs the derivative-free version of the Newton grid based algorithm
        for all test cases and non-parallel.
        """
        for caseName in benchmarkFunctions:
            gridRF = newtonGridFinder(
                caseName,
                numSamplePoints=50,
                derivativeFree=True,
                parallel=False,
            )
            reRan = testFunctions[caseName].reRan
            imRan = testFunctions[caseName].imRan
            gridRF.calculateRoots(reRan, imRan, precision=(5, 5))



"""
This module contains dataclasses for storing results fof command-line parsing.

Authors:\n
- Philipp Schuette\n
"""

from dataclasses import dataclass
from typing import Optional, Tuple


@dataclass
class SettingsParseResults:
    """
    Container for parsing results related to settings.
    """

    doPrint: bool
    container: str
    algorithm: str
    estimator: str
    logLevel: str
    verbose: str
    precision: Optional[Tuple[int, int]]


@dataclass
class PluginParseResults:
    """
    Container for parsing results related to plugins.
    """

    listPlugins: bool
    listModules: bool
    install: str
    uninstall: str


@dataclass
class InstallTestingParseResults:
    """
    Container for parsing results related to testing of the installation.
    """

    doTest: bool



"""
Module containing the `containers` used for root storage.

This module contains the abstract container interface `RootContainer` as well
as its concrete implementations.
"""

from pyzeal.utils.containers.root_container import RootContainer

__all__ = [
    "RootContainer",
]



"""
This module contains a collection of utility functions used in the tests of the
PyZEAL project.

Authors:\n
- Luca Wasmuth\n
- Philipp Schuette\n
"""

from typing import Tuple

from pyzeal.pyzeal_types.root_types import tVec


def rootsMatchClosely(
    roots1: tVec,
    roots2: tVec,
    precision: Tuple[int, int],
    allowSubset: bool = False,
) -> bool:
    """
    Determine if `roots1` and `roots2` contain the same roots, up to a number
    of decimal places given by `precision`.

    :param roots1: first vector of roots
    :param roots2: second vector of roots
    :param precision: number of decimal places where roots are considered equal
        the second is sufficient
    :return: flag indicating if `roots1` and `roots2` are the same
    """
    assert len(roots1) == len(roots2) or allowSubset

    deltaReal = 10 ** (-precision[0])
    deltaImag = 10 ** (-precision[1])

    for root1 in roots1:
        matchFound = False
        for root2 in roots2:
            if (
                abs(root1.real - root2.real) < deltaReal
                and abs(root1.imag - root2.imag) < deltaImag
            ):
                matchFound = True
                break
        if not matchFound:
            return False

    return True



"""
Class ParallelRootFinder from the package pyzeal.
This module defines an implementation of the main root finding API as defined
by the `RootFinderInterface` protocol. It differs from `RootFinder` in the fact
that it devides the search region into sub-regions and delegates these to a
number of appropriate algorithms working in parallel. The parallelism is
realized by using the standard library `multiprocessing` module.

Authors:\n
- Philipp Schuette\n
"""

from multiprocessing import Manager, Pool
from os import cpu_count, getpid
from signal import SIG_IGN, SIGINT, signal
from typing import List, Optional, Tuple, cast

from numpy import linspace
from rich.progress import TaskID

from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.pyzeal_types.parallel_types import FinderProgressManager, tQueue
from pyzeal.pyzeal_types.root_types import tHoloFunc
from pyzeal.rootfinders.rootfinder import RootFinder
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.filter_context import FilterContext
from pyzeal.utils.finder_progress import FinderProgressBar
from pyzeal.utils.root_context import RootContext


class ParallelRootFinder(RootFinder):
    """
    Parallel (multiprocessing) implementation of the main root finding API.
    """

    def __init__(
        self,
        f: tHoloFunc,
        df: Optional[tHoloFunc] = None,
        *,
        containerType: ContainerTypes = ContainerTypes.DEFAULT,
        algorithmType: AlgorithmTypes = AlgorithmTypes.DEFAULT,
        estimatorType: EstimatorTypes = EstimatorTypes.DEFAULT,
        precision: Optional[Tuple[int, int]] = None,
        numSamplePoints: Optional[int] = None,
        verbose: Optional[bool] = None,
    ) -> None:
        """
        Initialize a parallel (multiprocessing) root finder.

        :param f: the function whose roots should be calculated
        :param df: the derivative of `f`
        :param containerType: the type of container found roots are stored in
        :param algorithmType: the type of algorithm used for root finding
        :param precision: the accuracy at which roots are considered exact
        :param numSamplePoints: determines grid size for `NewtonGridAlgorithm`
        :param verbose: flag that toggles the command line progress bar
        """
        # every algorithm invocation uses numSamplePoints on its subgrid!
        numSamplePoints = (
            int(numSamplePoints / (cpu_count() or 1)) + 1
            if numSamplePoints
            else None
        )
        super().__init__(
            f=f,
            df=df,
            containerType=containerType,
            algorithmType=algorithmType,
            estimatorType=estimatorType,
            precision=precision,
            numSamplePoints=numSamplePoints,
            verbose=verbose,
        )

    def __str__(self) -> str:
        "Simple string representation of a `ParallelRootFinder`."
        if self.df is not None:
            return (
                "ParallelRootFinder(f="
                + f"{getattr(self.f, '__name__', '<unnamed>')}, "
                + f"df={getattr(self.df, '__name__', '<unnamed>')})"
            )
        return (
            "ParallelRootFinder(f="
            + f"{getattr(self.f, '__name__', '<unnamed>')}, "
            + "df=None)"
        )

    def calculateRoots(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        precision: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Parallel implementation of the root finding interface as defined in
        `RootFinderInterface`.

        :param reRan: horizontal extend of the complex region to search in
        :param imRan: vertical extend of the complex region to search in
        :param precision: accuracy of the search in real and imaginary parts
        """
        # if no precision was given, use default precision from constructor
        precision = precision or self.precision
        # if a rounding container is used we must calculate with an additional
        # digit of internal precision to obtain correct results after rounding
        precision = (precision[0] + 1, precision[1] + 1)
        # desymmetrize the input rectangle
        (x1, x2), (y1, y2) = self.desymmetrizeDomain(reRan, imRan, precision)

        # initialize a shared progress bar
        FinderProgressManager.register("finderProgress", FinderProgressBar)
        with FinderProgressManager() as manager:
            progress = manager.finderProgress() if self.verbose else None
            task: Optional[TaskID] = None
            if progress is not None:
                task = progress.addTask((x2 - x1) * (y2 - y1))
                progress.start()
                self.logger.debug("starting progress bar...")

            # construct a list of root contexts, several for each child process
            rootQueue: tQueue = cast(tQueue, Manager().Queue())
            contexts = self.createRootJobs(
                numProcesses=cpu_count() or 1,
                reRan=(x1, x2),
                imRan=(y1, y2),
                rootQueue=rootQueue,
                precision=precision,
                progress=progress,
                task=task,
            )

            with Pool(initializer=ParallelRootFinder.suppressSig) as pool:
                # shut down root search orderly upon command line signals
                try:
                    self.logger.info("attempting to calculate roots...")
                    pool.starmap(
                        self.rootWorker,
                        [(context,) for context in contexts],
                        chunksize=cpu_count() or 1,
                    )
                    if progress is not None and task is not None:
                        progress.update(
                            task, description=("[green] search finished!")
                        )
                    self.logger.debug("all child processes returned normally!")
                except KeyboardInterrupt:
                    self.logger.warning(
                        "calculation interrupted - some roots may be missing!"
                    )
                    if progress and task:
                        progress.stop_task(task)
                        progress.update(task, visible=False)
                        progress.refresh()
                if progress is not None and task is not None:
                    progress.stop()

            # add found roots to the current instance's container
            self.logger.debug("transferring roots from queue to container!")
            precision = (precision[0] - 1, precision[1] - 1)
            while not rootQueue.empty():
                self.container.addRoot(
                    rootQueue.get(),
                    FilterContext(self.f, (x1, x2), (y1, y2), precision),
                )
            self.logger.info("parallel root search finished!")

    def createRootJobs(
        self,
        numProcesses: int,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        rootQueue: tQueue,
        precision: Tuple[int, int],
        progress: Optional[FinderProgressBar],
        task: Optional[TaskID],
    ) -> List[RootContext]:
        """
        Convenience method that constructs a list of RootContext objects on
        which child processes operate by applying a concrete RootAlgorithm.

        :param numProcesses: The number of processes, which determines the
            amount of contexts that are returned
        :param reRan: Search range for the real part
        :param imRan: Search range for the imaginary part
        :param rootQueue: Queue in which new roots are added
        :param precision: accuracy of search in real and imaginary parts
        :param progress: Progress bar handle
        :param task: TaskID for the progress bar
        :return: List of RootContexts on which child processes can operate
        """
        self.logger.debug(
            "initializing context jobs for %d child processes...", numProcesses
        )
        realPts = linspace(reRan[0], reRan[1], numProcesses + 1)
        imagPts = linspace(imRan[0], imRan[1], numProcesses + 1)
        plainContainer = ContainerFactory.getConcreteContainer(
            ContainerTypes.PLAIN_CONTAINER, queue=rootQueue
        )
        contexts: List[RootContext] = []
        for i in range(len(realPts) - 1):
            for j in range(len(imagPts) - 1):
                contexts.append(
                    RootContext(
                        f=self.f,
                        df=self.df,
                        container=plainContainer,
                        reRan=(realPts[i], realPts[i + 1]),
                        imRan=(imagPts[j], imagPts[j + 1]),
                        precision=precision,
                        progress=progress,
                        task=task,
                    )
                )
        return contexts

    def rootWorker(self, context: RootContext) -> None:
        """
        Worker function that executes a root finding algorithm in a child
        process.

        :param context: Context object on which roots are searched
        """
        self.logger.info("starting root job in pid=%d!", getpid())
        self.algorithm.calcRoots(context)
        self.logger.info("finished root job in pid=%d!", getpid())

    @staticmethod
    def suppressSig() -> None:
        "Initialization routine setting workers to ignore `ctrl+c`."
        signal(SIGINT, SIG_IGN)



"""
This module provides a simple cache to store and retrieve intermediate
results during computation.

Authors:\n
- Philipp Schuette
"""

from logging import DEBUG
from typing import Dict, Optional, Tuple

from pyzeal.pyzeal_logging.loggable import Loggable


class EstimatorCache(Loggable):
    """
    A simple in-memory cache that can store and retrieve total argument changes
    along horizontal and vertical lines in the complex plane.
    """

    __slots__ = ("_cache", "cacheHits", "cacheMisses")

    def __init__(self) -> None:
        """
        Initializes a new `EstimatorCache`.
        """
        self._cache: Dict[int, Dict[Tuple[complex, complex], complex]] = {}
        self.logger.info("initialized a new argument estimator cache...")
        # cache hits and misses are only recorded if logging is set to DEBUG
        self.cacheHits = 0
        self.cacheMisses = 0

    def store(
        self,
        order: int,
        zStart: complex,
        zEnd: complex,
        argument: complex,
    ) -> None:
        """
        Store the total argument change associated with a horizontally or
        vertically oriented range of complex numbers.

        :param order: Order of the moment to be stored
        :param zStart: Starting point of the line
        :param zEnd: End point of the line
        :param argument: Total argument change
        """
        if (orderCache := self._cache.get(order, None)) is None:
            self._cache[order] = {(zStart, zEnd): argument}
        else:
            orderCache[(zStart, zEnd)] = argument
        self.logger.debug(
            "stored value %s under key %s in estimator cache!",
            str(argument),
            str((order, (zStart, zEnd))),
        )

    def retrieve(
        self,
        order: int,
        zStart: complex,
        zEnd: complex,
    ) -> Optional[complex]:
        """
        Retrieve the total argument change associated with a horizontally or
        vertically oriented range of complex numbers. Returns `None` if the
        requested entry is not present.

        :param order: Order of the moment to be retrieved
        :param zStart: Starting point of the line
        :param zEnd: End point of the line
        :return: Total argument change if the cache contains a value, else
            None is returned.
        """
        if (orderCache := self._cache.get(order, None)) is None:
            value = None
        else:
            value = orderCache.get((zStart, zEnd), None)
        self.logger.debug(
            "retrieved value %s under key %s from estimator cache!",
            str(value),
            str((zStart, zEnd)),
        )
        if self.logger.isEnabledFor(DEBUG):
            self.cacheHits += 1 if value else 0
            self.cacheMisses += 1 if not value else 0
        return value

    def remove(
        self,
        order: int,
        zStart: complex,
        zEnd: complex,
    ) -> None:
        """
        Remove the total argument change associated with a horizontally or
        vertically oriented range of complex numbers.

        :param order: Order of the moment to remove
        :param zStart: Starting point of the line
        :param zEnd: End point of the line
        """
        if order in self._cache:
            value = self._cache[order].pop((zStart, zEnd), None)
            self.logger.debug(
                "removed value %s under key %s from estimator cache!",
                str(value),
                str((zStart, zEnd)),
            )

    def dirty(self) -> bool:
        """
        Returns `True` if the cache contains anything.

        :return: `True` if the cache contains anything.
        """
        return len(self._cache) > 0

    def reset(self) -> None:
        """
        Resets the cache by clearing all stored values and resetting
        the hit and miss counters.
        """
        self._cache.clear()
        self.cacheHits = 0
        self.cacheMisses = 0



"""
This module contains tests for the plain container implementation.
"""

from typing import List

import pytest

from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.filter_context import FilterContext


def testAddPlainContainer(
    plainContainer: RootContainer, filterContexts: List[FilterContext]
) -> None:
    """
    Test the plain container `addRoot` implementation.
    """
    i = 0
    for filterContext in filterContexts:
        assert len(plainContainer.getRoots()) == 3 * i

        plainContainer.addRoot((1.23456789, 0), filterContext)
        assert len(plainContainer.getRoots()) == 3 * i + 1

        plainContainer.addRoot((1.2345678, -1), filterContext)
        assert len(plainContainer.getRoots()) == 3 * i + 2

        plainContainer.addRoot((-1.23456789, 2), filterContext)
        assert len(plainContainer.getRoots()) == 3 * i + 3

        i += 1

    assert (
        plainContainer.getRoots() == [1.23456789, 1.2345678, -1.23456789] * 3
    ).all()


def testRemovePlainContainer(
    plainContainer: RootContainer, filterContexts: List[FilterContext]
) -> None:
    """
    Test the plain container `removeRoot` implementation.
    """
    plainContainer.addRoot((0, 1), filterContexts[0])
    plainContainer.addRoot((0, 1), filterContexts[0])

    assert len(plainContainer.getRoots()) == 2
    assert not plainContainer.removeRoot((0, 1))
    assert len(plainContainer.getRoots()) == 2


def testClearPlainContainer(
    plainContainer: RootContainer, filterContexts: List[FilterContext]
) -> None:
    """
    Test the plain container `clear` implementation.
    """
    plainContainer.addRoot((0, 1), filterContexts[0])
    plainContainer.addRoot((0, 1), filterContexts[0])

    assert len(plainContainer.getRoots()) == 2
    with pytest.raises(NotImplementedError):
        plainContainer.clear()
    assert len(plainContainer.getRoots()) == 2


def testRegisterUnregisterFilterPlainContainer(
    plainContainer: RootContainer,
) -> None:
    """
    Test the plain container `registerFilter` implementation.
    """
    with pytest.raises(NotImplementedError):
        plainContainer.registerFilter(lambda r, c: False, "alwaysFalse")
    with pytest.raises(NotImplementedError):
        plainContainer.unregisterFilter("alwaysFalse")



"""
This module contains tests of the main root finding algorithms using
polynomials with roots generated by the `hypothesis` package.

Authors:\n
- Luca Wasmuth\n
"""

from datetime import timedelta
from typing import List

import numpy as np
import pytest
from hypothesis import Phase, given, settings, strategies
from numpy.polynomial import Polynomial

from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.pyzeal_types.filter_types import FilterTypes
from pyzeal.pyzeal_types.root_types import tHoloFunc
from pyzeal.rootfinders import RootFinder
from pyzeal.rootfinders.parallel_finder import ParallelRootFinder
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.utils import rootsMatchClosely
from pyzeal.utils.service_locator import ServiceLocator

settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)


@pytest.mark.slow
@pytest.mark.parametrize(
    "algorithm",
    [
        AlgorithmTypes.NEWTON_GRID,
        AlgorithmTypes.SIMPLE_ARGUMENT,
        AlgorithmTypes.SIMPLE_ARGUMENT_NEWTON,
    ],
)
@pytest.mark.parametrize(
    "estimator",
    [EstimatorTypes.SUMMATION_ESTIMATOR, EstimatorTypes.QUADRATURE_ESTIMATOR],
)
@pytest.mark.parametrize("parallel", [True, False])
@given(
    strategies.lists(
        strategies.complex_numbers(max_magnitude=5), min_size=2, max_size=5
    )
)
@settings(
    deadline=(timedelta(seconds=9)),
    max_examples=5,
    phases=[Phase.explicit, Phase.reuse, Phase.generate, Phase.explain],
)
def testRootFinderHypothesis(
    algorithm: AlgorithmTypes,
    estimator: EstimatorTypes,
    parallel: bool,
    roots: List[complex],
) -> None:
    """
    Test the root finder algorithm based on a simple partial integration of the
    classical argument principle combined with a Newton algorithm upon
    sufficient refinement of the subdivision into rectangles. The testfunctions
    are polynomials whose roots are generated automatically using the
    hypothesis package.

    :param algorithm: The type of algorithm to test
    :param estimator: The type of estimator to test
    :param parallel: Whether the tested finder works in parallel
    :param roots: Roots of the test polynomial
    """
    # We only find a higher-order zero once, so we have to remove duplicates
    uniqueRoots = list(set(np.round(roots, 2)))
    polynomial = Polynomial.fromroots(uniqueRoots)
    f: tHoloFunc = polynomial
    df: tHoloFunc = polynomial.deriv()
    Finder = RootFinder if parallel else ParallelRootFinder
    hrf = Finder(
        f,
        df,
        numSamplePoints=30,
        algorithmType=algorithm,
        containerType=ContainerTypes.ROUNDING_CONTAINER,
        estimatorType=estimator,
        verbose=False,
    )

    hrf.setRootFilter(filterType=FilterTypes.FUNCTION_VALUE_ZERO)
    hrf.setRootFilter(filterType=FilterTypes.ZERO_IN_BOUNDS)
    hrf.calculateRoots((-10.1, 10.2), (-10.3, 10.4), precision=(3, 3))

    foundRoots = np.sort_complex(hrf.roots)
    expectedRoots = np.sort_complex(np.array(uniqueRoots))

    assert rootsMatchClosely(
        foundRoots, expectedRoots, precision=(3, 3), allowSubset=True
    )



"""
This module provides a static factory class which maps the argument estimators
available in the `EstimatorType` enumeration to appropriate members of the
`ArgumentEstimator` type.

Authors:\n
- Philipp Schuette\n
"""

from typing import Optional

from pyzeal.algorithms.estimators import ArgumentEstimator, EstimatorCache
from pyzeal.algorithms.estimators.quad_estimator import QuadratureEstimator
from pyzeal.algorithms.estimators.sum_estimator import SummationEstimator
from pyzeal.pyzeal_logging.log_manager import LogManager
from pyzeal.pyzeal_logging.logger_facade import PyZEALLogger
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.service_locator import ServiceLocator


class EstimatorFactory:
    "Static factory class used to create instances of argument estimators."

    # module level logger
    _logger: Optional[PyZEALLogger] = None

    @staticmethod
    def getConcreteEstimator(
        estimatorType: EstimatorTypes,
        *,
        cache: EstimatorCache,
        numPts: Optional[int] = None,
        deltaPhi: Optional[float] = None,
        maxPrecision: Optional[float] = None,
    ) -> ArgumentEstimator:
        """
        Construct and return an estimator instance based on the given type
        of estimator `estimatorType`.

        :param estimatorType: Type of estimator to return
        :param numPts: Number of sampling points
        :param deltaPhi: Threshold for argument change in a single step
        :param maxPrecision: Maximum precision, after which no more refinement
            takes place
        :param cache: Cache to store argument changes
        :return: Estimator instance with given parameters
        """
        if EstimatorFactory._logger is None:
            EstimatorFactory._logger = LogManager.initLogger(
                __name__.rsplit(".", maxsplit=1)[-1],
                ServiceLocator.tryResolve(SettingsService).logLevel,
            )
        if estimatorType == EstimatorTypes.SUMMATION_ESTIMATOR:
            EstimatorFactory._logger.debug(
                "requested a new phase summation based argument estimator..."
            )
            if numPts is None or deltaPhi is None or maxPrecision is None:
                raise ValueError(
                    "SUMMATION_ESTIMATOR requires numPts, deltaPhi and \
                        maxPrecision!"
                )
            return SummationEstimator(
                numPts=numPts,
                deltaPhi=deltaPhi,
                maxPrecision=maxPrecision,
                cache=cache,
            )
        if estimatorType == EstimatorTypes.QUADRATURE_ESTIMATOR:
            EstimatorFactory._logger.debug(
                "requested a new quadrature based argument estimator..."
            )
            return QuadratureEstimator(cache=cache)

        EstimatorFactory._logger.debug(
            "requested a new default argument estimator..."
        )
        settings = ServiceLocator.tryResolve(SettingsService)
        return EstimatorFactory.getConcreteEstimator(
            settings.defaultEstimator,
            numPts=numPts,
            deltaPhi=deltaPhi,
            maxPrecision=maxPrecision,
            cache=cache,
        )



"""
Module cli_parser.py from the pyzeal_cli package.
This module implements a very PyZEAL specific command line argument parser
which gets used by the main entry point `__main__.py` to parse incoming user
arguments.

Authors:\n
- Philipp Schuette\n
"""

from __future__ import annotations

from argparse import ArgumentParser, _SubParsersAction
from importlib.metadata import version
from typing import Final, Tuple

from pyzeal.cli.parse_results import (
    InstallTestingParseResults,
    PluginParseResults,
    SettingsParseResults,
)
from pyzeal.cli.parser_facade import PyZEALParserInterface


class PyZEALParser(ArgumentParser, PyZEALParserInterface):
    """
    Specialized `argparse.ArgumentParser` which parses arguments provided to
    the PyZEAL CLI by a user.
    """

    MAIN_DESCRIPTION: Final[str] = (
        "welcome to the PyZEAL project! from the command line you can control "
        "various aspects of this package, like viewing and manipulating the "
        "settings which control its default behaviour."
    )
    SETTINGS_PLUGINS_DESCRIPTION: Final[str] = (
        "manipulate PyZEAL behaviour by changing the (default) settings and"
        " (un-)installing (custom) plugins"
    )
    PROGRAM_NAME: Final[str] = "pyzeal"
    VERSION: Final[str] = version(PROGRAM_NAME)

    def __init__(self) -> None:
        """
        Initialize a new `PyZEALParser` parser instance. The parser recognizes
        basic optional arguments like `--help` and `--version` as well as the
        subcommands `change` and `view` related to customization of settings.
        """
        super().__init__(
            description=PyZEALParser.MAIN_DESCRIPTION,
            prog=f"{PyZEALParser.PROGRAM_NAME}",
        )

        self.addVersionOption()
        # one could also let the user choose from a set of submodules to test
        self.addTestingOption()

        # add subcommands for options and plugins
        subParsers = self.add_subparsers(
            title=PyZEALParser.SETTINGS_PLUGINS_DESCRIPTION,
            help="view/change settings, (un-)install plugins and run tests",
            parser_class=ArgumentParser,
        )

        self.addViewSubcommand(subParsers)
        self.addChangeSubcommand(subParsers)
        self.addPluginSubcommand(subParsers)

    def addVersionOption(self) -> None:
        """
        Add version info option to the cli.
        """
        self.add_argument(
            "--version",
            action="version",
            version=f"%(prog)s {PyZEALParser.VERSION}",
        )

    def addTestingOption(self) -> None:
        """
        Add complete project testing option to the cli.
        """
        self.add_argument(
            "-t", "--test", action="store_true", help="test complete project"
        )

    def addViewSubcommand(
        self, subParsers: _SubParsersAction[ArgumentParser]
    ) -> None:
        """
        Add view subcommand and its options to the cli.

        :param subParsers: _description_
        """
        viewParser = subParsers.add_parser("view")
        viewParser.add_argument(
            "-p", "--print", action="store_true", help="print current settings"
        )

    def addChangeSubcommand(
        self, subParsers: _SubParsersAction[ArgumentParser]
    ) -> None:
        """
        Add change subcommand and its options to the cli.

        :param subParsers: _description_
        """
        changeParser = subParsers.add_parser("change")
        changeParser.add_argument(
            "--container",
            dest="container",
            choices=["rounding"],
            help="change current default container",
        )
        changeParser.add_argument(
            "--algorithm",
            choices=[
                "newton_grid",
                "simple_argument",
                "simple_argument_newton",
                "associated_polynomial",
            ],
            help="change current default algorithm",
        )
        changeParser.add_argument(
            "--estimator",
            choices=["summation", "quadrature"],
            help="change current default estimator",
        )
        changeParser.add_argument(
            "--log-level",
            choices=["debug", "info", "warning", "error", "critical"],
            help="change current default log level",
        )
        changeParser.add_argument(
            "--verbose",
            choices=["true", "True", "false", "False"],
            help="change current default verbosity level",
        )
        changeParser.add_argument(
            "--precision",
            nargs=2,
            type=int,
            help="change current default root finding precision",
        )

    def addPluginSubcommand(
        self, subParsers: _SubParsersAction[ArgumentParser]
    ) -> None:
        """
        Add plugin subcommand and its options to the cli.

        :param subParsers: The subparser to add Plugin commands to
        """
        pluginParser = subParsers.add_parser("plugin")
        pluginParser.add_argument(
            "-l", "--list", action="store_true", help="list installed plugins"
        )
        pluginParser.add_argument(
            "-m",
            "--modules",
            action="store_true",
            help="list contents of installation directory",
        )
        pluginParser.add_argument(
            "--install",
            help="install a given file containing a plugin (or plugin data)",
        )
        pluginParser.add_argument(
            "--uninstall",
            help="uninstall a given plugin-related (data or source) file",
        )

    def parseArgs(
        self,
    ) -> Tuple[
        SettingsParseResults, PluginParseResults, InstallTestingParseResults
    ]:
        """
        Read command line arguments, parse the read arguments and return them
        wrapped according to the `pyzeal_cli` data contract for parsed command
        line arguments.

        :return: the wrapped results of the parsing process
        """
        # fetch cli arguments
        args = super().parse_args()

        # return wrapped cli arguments
        parseArgs = SettingsParseResults(
            doPrint=getattr(args, "print", False),
            container=getattr(args, "container", ""),
            algorithm=getattr(args, "algorithm", ""),
            estimator=getattr(args, "estimator", ""),
            logLevel=getattr(args, "log_level", ""),
            verbose=getattr(args, "verbose", ""),
            precision=tuple(precision)  # type: ignore
            if (precision := getattr(args, "precision", None))
            else None,
        )
        pluginArgs = PluginParseResults(
            listPlugins=getattr(args, "list", False),
            listModules=getattr(args, "modules", False),
            install=getattr(args, "install", ""),
            uninstall=getattr(args, "uninstall", ""),
        )
        testingArgs = InstallTestingParseResults(
            doTest=getattr(args, "test", False)
        )

        return parseArgs, pluginArgs, testingArgs



"""
This module contains tests of the grid-based Newton algorithm implementation
of the root finding algorithm interface.

Authors:\n
- Luca Wasmuth\n
"""

import numpy as np
import pytest

from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.finder_helpers import newtonGridFinder
from pyzeal.tests.resources.finder_test_cases import testFunctions
from pyzeal.tests.resources.utils import rootsMatchClosely
from pyzeal.utils.service_locator import ServiceLocator

settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)

# some test functions do not work due to algorithmic limitations
KNOWN_FAILURES = ["x^30", "x^50", "x^100", "1e6 * x^100"]


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("testName", sorted(testFunctions.keys()))
@pytest.mark.parametrize("parallel", [False, True])
def testNewtonGridRootFinder(testName: str, parallel: bool) -> None:
    """
    Test the parallel and serial implementations of the `RootFinderInterface`
    using the `NewtonGridAlgorithm` implementation of the `FinderAlgorithm`.

    :param testName: Name of the test case
    :param parallel: If roots should be searched in parallel
    """
    if testName in KNOWN_FAILURES:
        pytest.skip()

    gridRF = newtonGridFinder(testName, numSamplePoints=30, parallel=parallel)
    reRan = testFunctions[testName].reRan
    imRan = testFunctions[testName].imRan
    precision = testFunctions[testName].precision
    if testName == "exp(x)-1":
        precision = (2, 2)
    gridRF.calculateRoots(reRan, imRan, precision=precision)
    foundRoots = np.sort_complex(gridRF.roots)
    expectedRoots = np.array(testFunctions[testName].expectedRoots)

    assert rootsMatchClosely(foundRoots, expectedRoots, precision=precision)



"""
This module provides the main CLI entry point of the PyZEAL project through the
function `mainPyZEAL`. At the moment it provides facilities to query the
currently installed PyZEAL version, view and manipulate settings,
(un-)install plugins, and invoke unit tests on the local installation.

Authors:\n
- Philipp Schuette\n
"""

from sys import argv

from pyzeal.cli.controller_facade import CLIControllerFacade
from pyzeal.cli.parser_facade import PyZEALParserInterface
from pyzeal.pyzeal_types.init_modes import InitModes
from pyzeal.utils.initialization_handler import PyZEALInitializationHandler
from pyzeal.utils.service_locator import ServiceLocator


class PyZEALEntry:
    """
    Class serving as the main entry point to the `PyZEAL` command line
    interface.
    """

    @staticmethod
    def mainPyZEAL() -> None:
        """
        Main entry point for the CLI of the PyZEAL project.
        """
        PyZEALInitializationHandler.initPyZEALServices(InitModes.CLI)

        parser = ServiceLocator.tryResolve(PyZEALParserInterface)
        settingsArgs, pluginArgs, testingArgs = parser.parseArgs()

        # check if any arguments were provided and respond with usage hint
        if len(argv) < 2:
            print("this is the CLI of the PyZEAL package. use '-h' for help.")

        controller = ServiceLocator.tryResolve(CLIControllerFacade)
        controller.handleViewSubcommand(settingsArgs)
        controller.handleChangeSubcommand(settingsArgs)
        controller.handlePluginSubcommand(pluginArgs)
        optionSelected = controller.handleTestingOption(testingArgs)

        # a valid subcommand was selected but with no meaningful options
        if len(argv) == 2 and not optionSelected:
            print("use '-h' with your subcommand for help.")



"""
Module constants.py from the package `pyzeal.algorithms.estimators`.

This module collects constants used in various argument estimators for ease of
reference and consistency.

Authors:\n
- Philipp Schuette\n
"""

from typing import Final

# number of sample points for integration
EXP_SAMPLE_POINTS: Final[int] = 10
# maximal sample points for integration
MAX_SAMPLE_POINTS: Final[int] = 21

# constant determining the refinement of complex arrays for large phi values
Z_REFINE: Final[int] = 100
# constant determining the maximal length of z-arrays
MAX_Z_LENGTH: Final[int] = 100



"""
TODO

Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol, runtime_checkable

import pytest

from pyzeal.utils.configuration_exception import InvalidServiceConfiguration
from pyzeal.utils.service_locator import ServiceLocator


# some classes for testing
@runtime_checkable
class FooInterface(Protocol):
    "Interface for test class."

    def method(self) -> None:
        "This does nothing."
        ...


class Foo(FooInterface):
    "Test class."

    def method(self) -> None:
        "Interface implementation."


@pytest.mark.locator
def testRegisterSingleton() -> None:
    "Test singleton registration."
    ServiceLocator.clearConfigurations()
    instance: FooInterface = Foo()

    # try valid registration
    assert ServiceLocator is ServiceLocator.registerAsSingleton(
        FooInterface, instance
    )
    resolvent = ServiceLocator.tryResolve(FooInterface)

    assert resolvent is instance

    # try invalid registration
    with pytest.raises(InvalidServiceConfiguration):
        ServiceLocator.registerAsSingleton(
            FooInterface, 0
        ).registerAsSingleton(FooInterface, Foo())


@pytest.mark.locator
def testRegisterTransient() -> None:
    "Test transient registration."
    ServiceLocator.clearConfigurations()

    # try valid registration
    assert ServiceLocator is ServiceLocator.registerAsTransient(
        FooInterface, Foo
    )
    resolvent1 = ServiceLocator.tryResolve(FooInterface)
    resolvent2 = ServiceLocator.tryResolve(FooInterface)

    assert isinstance(resolvent1, FooInterface)
    assert isinstance(resolvent1, Foo)
    assert isinstance(resolvent2, FooInterface)
    assert isinstance(resolvent2, Foo)

    # check that genuinely new objects were resolved
    assert resolvent1 is not resolvent2


@pytest.mark.locator
def testLocatorSealing() -> None:
    "Test sealing operation on locator."
    ServiceLocator.clearConfigurations()
    ServiceLocator.seal()

    # try valid registration on sealed locator
    with pytest.raises(ValueError):
        ServiceLocator.registerAsSingleton(FooInterface, Foo())

    # try valid registration on sealed locator
    with pytest.raises(ValueError):
        ServiceLocator.registerAsTransient(FooInterface, Foo)

    # try sealing locator twice
    with pytest.raises(ValueError):
        ServiceLocator.seal()


@pytest.mark.locator
def testInvalidConfiguration() -> None:
    "Test if invalid registered configurations raise exceptions."
    ServiceLocator.clearConfigurations()

    # try to resolve a non-existing configuration
    with pytest.raises(InvalidServiceConfiguration):
        ServiceLocator.tryResolve(FooInterface)

    # register an invalid configuration and try to resolve it
    ServiceLocator.registerAsTransient(
        FooInterface, lambda: None  # type: ignore
    )
    with pytest.raises(InvalidServiceConfiguration):
        ServiceLocator.tryResolve(FooInterface)



"""
Testing fixtures for the `RootContainer` tests.
"""

from typing import List

import pytest

from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.filter_context import FilterContext


@pytest.fixture(name="filterContexts")
def fixtureFilterContexts() -> List[FilterContext]:
    "Fixture providing a list of `FilterContext` instances."
    fContext1 = FilterContext(
        lambda x: (x - 1.23456789) * (x + 1.23456789), (-5, 5), (-5, 5), (3, 3)
    )
    fContext2 = FilterContext(lambda x: x + 1, (-1, 2), (-3, 4), (5, 6))
    return [fContext1, fContext2, fContext1]


@pytest.fixture(name="plainContainer")
def fixturePlainContainer() -> RootContainer:
    "Fixture providing a `PlainContainer` instance."
    container = ContainerFactory.getConcreteContainer(
        containerType=ContainerTypes.PLAIN_CONTAINER
    )
    return container


@pytest.fixture(name="roundingContainer")
def fixtureRoundingContainer() -> RootContainer:
    "Fixture providing a `PlainContainer` instance."
    container = ContainerFactory.getConcreteContainer(
        containerType=ContainerTypes.ROUNDING_CONTAINER
    )
    return container



"""
This module contains resources for estimator tests: A utility function to
generate `RootContext`s as well as test cases for argument estimation on a
rectangular contour and along a straight line.
"""

from typing import Dict, Tuple, cast

import numpy as np

from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.root_types import tHoloFunc, tVec
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.root_context import RootContext
from pyzeal.utils.service_locator import ServiceLocator

ServiceLocator.registerAsTransient(SettingsService, RAMSettingsService)


functions: dict[str, Tuple[tHoloFunc, tHoloFunc]] = {
    "2x^2+3": (lambda z: 2 * z**2 + 3, lambda z: 4 * z),
    "sin(x)": (
        lambda z: cast(tVec, -2.4 * np.sin(z)),
        lambda z: cast(tVec, -2.4 * np.cos(z)),
    ),
    "exp(ix)": (
        lambda z: cast(tVec, 5.1j * np.exp(z * 1j)),
        lambda z: cast(tVec, 5.1j * np.exp(z * 1j) * 1j),
    ),
}


def generateRootContext(
    name: str, reRan: Tuple[float, float], imRan: Tuple[float, float]
) -> RootContext:
    """
    Generate a `RootContext` for the test case `name` with given search ranges.

    :param name: Test case name
    :param reRan: Real search range
    :param imRan: Imaginary search range
    :return: Generated `RootContext`
    """
    return RootContext(
        f=functions[name][0],
        df=functions[name][1],
        container=ContainerFactory.getConcreteContainer(
            containerType=ContainerTypes.ROUNDING_CONTAINER
        ),
        reRan=reRan,
        imRan=imRan,
        precision=(5, 5),
    )


rectangleCases: Dict[str, Tuple[RootContext, int, float]] = {
    "2x^2+3,(-1,1)x(-1,3)": (
        generateRootContext("2x^2+3", (-1, 1), (-1, 3)),
        0,
        2 * np.pi,
    ),
    "sin(x),(-5,5)x(-5,5)": (
        generateRootContext("sin(x)", (-5, 5), (-5, 5)),
        0,
        6 * np.pi,
    ),
    "exp(ix),(-1,2)x(-3,4)": (
        generateRootContext("exp(ix)", (-1, 2), (-3, 4)),
        0,
        0.0,
    ),
}

lineCases: Dict[str, Tuple[RootContext, int, complex, complex, complex]] = {
    "2x^2+3,-1+1j:1+1j": (
        generateRootContext("2x^2+3", (-1, 1), (-1, 1)),
        0,
        -1 + 1j,
        1 + 1j,
        -1j * np.log((3 + 4j) / (3 - 4j)),
    ),
    "2x^2+3,-1+1j:-1-1j": (
        generateRootContext("2x^2+3", (-1, 1), (-1, 1)),
        0,
        -1 + 1j,
        -1 - 1j,
        1j * np.log((-3 + 4j) / (-3 - 4j)),
    ),
    "sin(x),1j:3j": (
        generateRootContext("sin(x)", (-1, 1), (1, 3)),
        0,
        0 + 1j,
        0 + 3j,
        -1j * np.log(np.sinh(3) / np.sinh(1)),
    ),
    "sin(x),1:3": (
        generateRootContext("sin(x)", (1, 3), (-1, 1)),
        0,
        1 + 0j,
        3 + 0j,
        -1j * np.log(np.sin(3) / np.sin(1)),
    ),
    "exp(ix),-1-1j:-1+1j": (
        generateRootContext("exp(ix)", (-1, 1), (-1, 1)),
        0,
        -1 - 1j,
        -1 + 1j,
        2.0j,
    ),
    "exp(ix),-1-1j:1-1j": (
        generateRootContext("exp(ix)", (-1, 1), (-1, 1)),
        0,
        -1 - 1j,
        1 - 1j,
        2.0,
    ),
}



"""
This module contains the base interface for plugins. Its methods provide the
necessary functions for plugin handling.

Authors:\n
- Philipp Schuette\n
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Generic, Tuple, Type, TypeVar, Union

from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.finder_algorithm import FinderAlgorithm
from pyzeal.cli.parser_facade import PyZEALParserInterface
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.containers.root_container import RootContainer

tPluggable = Union[
    FinderAlgorithm,
    ArgumentEstimator,
    RootContainer,
    SettingsService,
    PyZEALParserInterface,
]
T_co = TypeVar("T_co", bound=tPluggable, covariant=True)


class PyZEALPlugin(ABC, Generic[T_co]):
    """
    Interface for plugins. Methods for instantiation and plugin information
    need to be implemented.
    """

    def __str__(self) -> str:
        "Print a human-readable representation of the plugin."
        major, minor, patch = self.pluginVersion
        maxLength = 30
        cutoff = min(len(self.pluginName), maxLength)
        return (
            f"{self.pluginName[:cutoff]: <{maxLength}} "
            f" @ v{major}.{minor}.{patch}"
        )

    @staticmethod
    @abstractmethod
    def initialize() -> Callable[..., T_co]:
        """
        Returns a factory for the plugin's implementation of `T_co`.

        :return: Factory for plugin implementation of `T_co`.
        """

    @staticmethod
    @abstractmethod
    def getInstance() -> PyZEALPlugin[T_co]:
        """
        Return the singleton instance of `PyZEALPlugin` which provides a
        service of type `T_co`.

        :return: `PyZEALPlugin` instance.
        """

    @property
    @abstractmethod
    def pluginType(self) -> Type[T_co]:
        """
        Return the type provided by the plugin.

        :return: Service type of plugin.
        """
        ...

    @property
    @abstractmethod
    def pluginName(self) -> str:
        """
        Return the name of the plugin.

        :return: Plugin name.
        """
        ...

    @property
    @abstractmethod
    def pluginVersion(self) -> Tuple[int, int, int]:
        """
        Return the version of the plugin as (`major`, `minor`, `patch`).

        :return: Plugin version.
        """
        ...



"""
Provide rootfinder setup methods with common settings for testing purposes.
"""


from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.pyzeal_types.filter_types import FilterTypes
from pyzeal.rootfinders import (
    ParallelRootFinder,
    RootFinder,
    RootFinderInterface,
)
from pyzeal.tests.resources.finder_test_cases import testFunctions


def buildFinder(
    testName: str,
    parallel: bool,
    algorithmType: AlgorithmTypes,
    estimatorType: EstimatorTypes,
    numSamplePoints: int = 20,
    derivativeFree: bool = False,
) -> RootFinderInterface:
    """
    Build a `RootFinderInterface` instance from a given configuration.

    :param testName: Test case name
    :param parallel: Set to `True` if the RootFinder should search in
        parallel
    :param algorithmType: The type of algorithm to use within the finder
    :param estimatorType: The type of estimator to use with algorithms based on
        the argument principle
    :param numSamplePoints: `numSamplePoints` passed to the Newton grid
        algorithm
    :param derivativeFree: Set to `True` if the RootFinder should operate
        without the derivative
    :return: Initialized root finder instance
    """
    f = testFunctions[testName].testFunc
    df = None if derivativeFree else testFunctions[testName].testFuncDerivative
    Finder = ParallelRootFinder if parallel else RootFinder
    finder: RootFinderInterface = Finder(
        f,
        df,
        numSamplePoints=numSamplePoints,
        containerType=ContainerTypes.ROUNDING_CONTAINER,
        algorithmType=algorithmType,
        estimatorType=estimatorType,
    )
    finder.setRootFilter(filterType=FilterTypes.FUNCTION_VALUE_ZERO)
    finder.setRootFilter(filterType=FilterTypes.ZERO_IN_BOUNDS)

    return finder


def newtonGridFinder(
    testName: str,
    numSamplePoints: int = 20,
    parallel: bool = False,
    derivativeFree: bool = False,
) -> RootFinderInterface:
    """
    Returns a NEWTON_GRID RootFinder for the test case `testName`

    :param testName: Test case name
    :param numSamplePoints: `numSamplePoints` passed to the RootFinder,
        defaults to 20
    :param parallel: Set to `True` if the RootFinder should search in
        parallel, defaults to False
    :param derivativeFree: Set to `True` if the RootFinder should operate
        without the derivative, defaults to False
    :return: Initialized root finder instance
    """
    gridFinder = buildFinder(
        testName=testName,
        numSamplePoints=numSamplePoints,
        parallel=parallel,
        derivativeFree=derivativeFree,
        algorithmType=AlgorithmTypes.NEWTON_GRID,
        estimatorType=EstimatorTypes.DEFAULT,
    )

    return gridFinder


def simpleArgumentRootFinder(
    testName: str,
    parallel: bool = False,
    estimatorType: EstimatorTypes = EstimatorTypes.SUMMATION_ESTIMATOR,
) -> RootFinderInterface:
    """
    Returns a root finder using the `SIMPLE_ARGUMENT` algorithm for the test
    case `testName`.

    :param testName: Test case name
    :param parallel: Set to `True` if the RootFinder should search in parallel,
        defaults to False
    :param estimatorType: The type of estimator to use with the simple argument
        algorithm
    :return: Initialized root finder instance
    """
    simpleFinder = buildFinder(
        testName=testName,
        parallel=parallel,
        algorithmType=AlgorithmTypes.SIMPLE_ARGUMENT,
        estimatorType=estimatorType,
    )

    return simpleFinder


def simpleArgumentNewtonRootFinder(
    testName: str,
    parallel: bool = False,
    estimatorType: EstimatorTypes = EstimatorTypes.SUMMATION_ESTIMATOR,
) -> RootFinderInterface:
    """
    Returns a root finder using the `SIMPLE_ARGUMENT_NEWTON` algorithm for the
    test case `testName`.

    :param testName: Test case name
    :param parallel: Set to `True` if the RootFinder should search in parallel,
        defaults to False
    :param estimatorType: The type of estimator to use with the simple argument
        with Newton method algorithm
    :return: Initialized root finder instance
    """
    simpleNewtonFinder = buildFinder(
        testName=testName,
        parallel=parallel,
        algorithmType=AlgorithmTypes.SIMPLE_ARGUMENT_NEWTON,
        estimatorType=estimatorType,
    )

    return simpleNewtonFinder



"""
Timing benchmark suite for the simple argument (with and without numerical
quadrature) with Newton algorithm based root finding algorithm.

Authors:\n
- Luca Wasmuth\n
- Philipp Schuette\n
"""

from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.finder_helpers import (
    simpleArgumentNewtonRootFinder,
)
from pyzeal.tests.resources.finder_test_cases import testFunctions
from pyzeal.utils.service_locator import ServiceLocator

benchmarkFunctions = [
    "x^2-1",
    "x^2+1",
    "x^4-1",
    "x^5-4x+2",
    "sin(x)",
    "exp(x)",
]

# disable progress bar
settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)


class SimpleArgumentNewtonSuite:
    """
    Timing benchmarks for the simple argument principle based root finder with
    additional Newton algorithm once sufficient precision has been attained.
    Includes simple polynomials as well as trigonometric and exponential
    functions.
    """

    def TimeSimpleArgumentNewtonSummation(self) -> None:
        """
        Runs the simple argument version of the rootfinding algorithm with
        summation based argument estimation and non-parallel.
        """
        for caseName in benchmarkFunctions:
            hrf = simpleArgumentNewtonRootFinder(
                caseName,
                parallel=False,
                estimatorType=EstimatorTypes.SUMMATION_ESTIMATOR,
            )
            reRan = testFunctions[caseName].reRan
            imRan = testFunctions[caseName].imRan
            hrf.calculateRoots(reRan, imRan, precision=(5, 5))

    def TimeSimpleArgumentNewtonQuadrature(self) -> None:
        """
        Analogous to `TimeSimpleArgumentSummation` but here the algorithm uses
        quadrature based argument estimation.
        """
        for caseName in benchmarkFunctions:
            hrf = simpleArgumentNewtonRootFinder(
                caseName,
                parallel=False,
                estimatorType=EstimatorTypes.QUADRATURE_ESTIMATOR,
            )
            reRan = testFunctions[caseName].reRan
            imRan = testFunctions[caseName].imRan
            hrf.calculateRoots(reRan, imRan, precision=(5, 5))



"""
TODO

Authors:\n
- Philipp Schuette\n
"""

from os import chdir
from os.path import dirname, join, realpath
from typing import Union

import pytest

from pyzeal.utils.install_test_facade import InstallTestingHandlerFacade


class InstallTestingHandler(InstallTestingHandlerFacade):
    "Concrete implementation of the `TestingHandlerFacade` interface."

    def __init__(self) -> None:
        "Initialize a new `TestingHandler` instance."

    # docstr-coverage:inherited
    def testModule(self, module: str) -> Union[int, pytest.ExitCode]:
        # change to the directory just above the tests/ directory
        chdir(dirname(realpath(__file__)))
        chdir("..")
        # obtain the full path to the module under test
        path = join("tests", module)
        # test the given module
        return pytest.main([path])



"""
Logging framework of the `PyZEAL` project.

This module exposes a simple logging facade and a `LogManager` which should be
used for any project-internal logging, e.g. in algorithms, root finders, or
plugins.
"""

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_logging.logger_facade import PyZEALLogger

__all__ = [
    "LogLevel",
    "PyZEALLogger",
]



"""
Module plugin_loader.py from the package PyZEAL.
This module handles discovering, loading and registering of plugins.

Authors:\n
- Philipp Schuette\n
"""

from __future__ import annotations

from abc import ABCMeta
from importlib import import_module
from os import listdir
from os.path import dirname, join, splitext
from re import compile as compileRegex
from types import ModuleType
from typing import Final, List, Optional, Type

from pyzeal.plugins.pyzeal_plugin import PyZEALPlugin, tPluggable
from pyzeal.pyzeal_logging.loggable import Loggable
from pyzeal.utils.service_locator import ServiceLocator

# default location where custom plugins are installed
PLUGIN_INSTALL_DIR: Final[str] = join(dirname(__file__), "custom_plugins")


class PluginLoader(Loggable):
    """
    This class handles plugin discovery, loading and registration.
    """

    _instance: Optional[PluginLoader] = None

    @staticmethod
    def getInstance() -> PluginLoader:
        """
        Return the global `PluginLoader` instance. If no instance exists,
        a new one is created and returned.

        :return: `PluginLoader` singleton instance.
        """
        return (
            PluginLoader._instance
            if PluginLoader._instance is not None
            else PluginLoader()
        )

    @staticmethod
    def loadPlugins(
        path: str = PLUGIN_INSTALL_DIR,
    ) -> List[PyZEALPlugin[tPluggable]]:
        """
        Load plugins present in `path`.

        :param path: Path to search for plugins, defaults to PLUGIN_INSTALL_DIR
        :return: List of loaded plugins.
        """
        instance = PluginLoader.getInstance()
        plugins: List[PyZEALPlugin[tPluggable]] = []
        for plugin in instance.locateAndLoadPlugins(path):
            pluginInstance = plugin.getInstance()
            ServiceLocator.registerAsTransient(
                pluginInstance.pluginType, plugin.initialize()
            )
            plugins.append(pluginInstance)
        return plugins

    def locateAndLoadPlugins(
        self, path: str = PLUGIN_INSTALL_DIR
    ) -> List[Type[PyZEALPlugin[tPluggable]]]:
        """
        Discover plugins at a given path and load them.

        :param path: Path to search for plugins, defaults to PLUGIN_INSTALL_DIR
        :return: List of found plugins.
        """
        plugins: List[Type[PyZEALPlugin[tPluggable]]] = []
        self.logger.info("starting plugin discovery in %s...", path)
        candidates = PluginLoader.discoverModules(path)
        self.logger.debug("contents of plugin directory: %s", str(candidates))
        for candidate in candidates:
            if not (candidate := PluginLoader.isPyFile(candidate)):
                continue
            self.logger.info("module %s might contain a plugin...", candidate)
            module = import_module(
                candidate, package="pyzeal_plugins.custom_plugins"
            )
            plugin = self.loadPlugin(module)
            if plugin is not None:
                plugins.append(plugin)
        return plugins

    def loadPlugin(
        self, candidateModule: ModuleType
    ) -> Optional[Type[PyZEALPlugin[tPluggable]]]:
        """
        Try to load a candidate plugin and return it if successful.

        :param candidateModule: Candidate plugin module
        :return: Plugin if an implementation has been found, else `None`.
        """
        attributeNames = PluginLoader.discoverAttributes(candidateModule)
        self.logger.debug(
            "module attributes %s found during plugin discovery!",
            str(attributeNames),
        )
        for attributeName in attributeNames:
            attribute = getattr(candidateModule, attributeName)
            if attribute == PyZEALPlugin:
                continue
            if isinstance(attribute, ABCMeta):
                if issubclass(attribute, PyZEALPlugin):
                    self.logger.info(
                        "plugin implementation [ %s ] found!",
                        str(attribute),
                    )
                    return attribute
        return None

    @staticmethod
    def isPyFile(filename: str) -> str:
        """
        Returns `True` if `filename` corresponds to a python file.

        :param filename: File to evaluate
        :return: `True` if `filename` corresponds to a python file.
        """
        if filename == "__init__.py":
            return ""

        name, extension = splitext(filename)
        return "." + name if extension.lower() == ".py" else ""

    @staticmethod
    def discoverModules(path: str = PLUGIN_INSTALL_DIR) -> List[str]:
        """
        Discover all possible plugin files in `path`. Ignore files starting
        with `__`.

        :param path: Path to search, defaults to PLUGIN_INSTALL_DIR
        :return: List of plugin candidates.
        """
        regex = compileRegex(r"__.*")
        candidates = [c for c in listdir(path) if not regex.match(c)]
        return candidates

    @staticmethod
    def discoverAttributes(candidateModule: ModuleType) -> List[str]:
        """
        Discover attributes of a candidate module. Ignores attributes
        starting with `__`.

        :param candidateModule: Candidate module
        :return: List of attributes.
        """
        regex = compileRegex(r"__.*")
        candidates = [c for c in dir(candidateModule) if not regex.match(c)]
        return candidates



"""
Module `root_types` from the package `pyzeal.pyzeal_types`.

This module provides common type aliases used throughout the PyZEAL project.

Authors:\n
- Philipp Schuette\n
"""

from typing import Callable, Tuple

import numpy as np
from numpy.typing import NDArray
from typing_extensions import TypeAlias

# general purpose type for numpy arrays with complex entries
tVec: TypeAlias = NDArray[np.complex128]

# type of functions our root finding algorithms can handle
tHoloFunc: TypeAlias = Callable[[tVec], tVec]

# type used to identify roots of holomorphic functions (point in the plane with
# its multiplicity)
tRoot: TypeAlias = Tuple[complex, int]

# type of rectangular grid used internally in simple argument rootfinders
tRecGrid: TypeAlias = Tuple[tVec, tVec, tVec, tVec]



"""
This module contains a base class for argument estimators, which should
inherit from this class and supply a concrete implementation.

Authors:\n
- Philipp Schuette\n
"""

from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np

from pyzeal.algorithms.estimators.estimator_cache import EstimatorCache
from pyzeal.pyzeal_logging.loggable import Loggable
from pyzeal.pyzeal_types.root_types import tVec
from pyzeal.utils.root_context import RootContext


class ArgumentEstimator(ABC, Loggable):
    """
    Interface for argument estimators. `calcMoment` us implemented here,
    while the concrete method of evaluation is determined by overriding
    `calcMomentAlongLine`.
    """

    def calcMoment(
        self,
        order: int,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        context: RootContext,
    ) -> complex:
        """
        Calculate the `order`-th moment of the logarithmic derivative of the
        target function `context.f` along the boundary of the rectangle
        specified by `reRan` x `imRan`. The concrete method by which the
        integral of the logarithm derivative gets calculated is determined by
        overriding `calcMomentAlongLine`.

        :param order: Order of the moment to be calculated.
        :param reRan: Interval describing the real part of the rectangle
        :param imRan: Interval describing the imaginary part of the rectangle
        :param context: `RootContext` containing the necessary information.
        :return: `order`-th moment of the logarithmic derivative of
            `context.f` along the boundary of the specified rectangle.
        """
        x1, x2 = reRan
        y1, y2 = imRan
        self.logger.debug(
            "estimating argument for rectangle [%f, %f] x [%f, %f]!",
            x1,
            x2,
            y1,
            y2,
        )
        phi: complex = 0

        # check if the requested complex line already resides in cache
        arguments: List[Tuple[complex, complex]] = [
            (x1 + y1 * 1j, x2 + y1 * 1j),
            (x2 + y1 * 1j, x2 + y2 * 1j),
            (x2 + y2 * 1j, x1 + y2 * 1j),
            (x1 + y2 * 1j, x1 + y1 * 1j),
        ]
        for zStart, zEnd in arguments:
            if entry := self.cache.retrieve(order, zStart, zEnd):
                phi += entry
            else:
                deltaPhi = self.calcMomentAlongLine(
                    order, zStart, zEnd, context
                )
                # store the missing entry in the cache
                self.cache.store(order, zStart, zEnd, deltaPhi)
                self.cache.store(order, zEnd, zStart, -deltaPhi)
                phi += deltaPhi

        self.logger.debug("estimated argument is %s", str(phi / (2.0 * np.pi)))
        return phi

    @abstractmethod
    def calcMomentAlongLine(
        self,
        order: int,
        zStart: complex,
        zEnd: complex,
        context: RootContext,
    ) -> complex:
        """
        Calculate the `order`-th moment of the logarithmic derivative of
        the target function `context.f` along the line given by `zStart` and
        `zEnd`.

        :param order: Order of the moment to calculate
        :param zStart: Starting point of the line
        :param zEnd: End point of the line
        :param context: `RootContext` containing the necessary information.
        :return: The moment as calculated along the given line.
        """

    @property
    @abstractmethod
    def cache(self) -> EstimatorCache:
        """
        Returns the cache used by this argument estimator.

        :return: Cache used by this argument estimator.
        """

    def genFuncArr(
        self, zStart: complex, zEnd: complex, context: RootContext, size: int
    ) -> Tuple[tVec, tVec]:
        """
        TODO.
        """
        pos = "horizontal" if zStart.imag == zEnd.imag else "vertical"
        zArr = np.linspace(zStart, zEnd, size)
        funcArr = context.f(zArr)
        zerosOnLine = np.where(funcArr == 0)[0]

        while zerosOnLine.size > 0:
            self.logger.debug(
                "simple argument found %d roots on the line [%s, %s]",
                zerosOnLine.size,
                str(zArr[0]),
                str(zArr[-1]),
            )
            # order of these zeros is not determined further, so put 0
            for newRoot in zArr[zerosOnLine]:
                context.container.addRoot(
                    (newRoot, 0), context.toFilterContext()
                )
            # adjust line according to 'pos'
            if pos == "vertical":
                zArr += 2 * 10 ** (-context.precision[0])
            else:
                zArr += 2j * 10 ** (-context.precision[1])
            funcArr = context.f(zArr)
            zerosOnLine = np.where(funcArr == 0)[0]

        return zArr, funcArr



"""
Class `SettingsService` from the package `pyzeal_settings`.
This module defines a protocol for a generic settings provider.

Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol, Tuple, runtime_checkable

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.settings.core_settings_service import CoreSettingsService


@runtime_checkable
class SettingsService(CoreSettingsService, Protocol):
    """
    Class providing a layer of abstraction for storage and retrieval of PyZEAL
    related settings. Concrete `SettingService` implementations can choose
    freely their data model and persistence layer.
    """

    @property
    def logLevel(self) -> LogLevel:
        """
        Get the current LogLevel setting.

        :return: Current LogLevel
        """
        ...

    @logLevel.setter
    def logLevel(self, value: LogLevel) -> None:
        """
        Set the LogLevel to `value`.

        :param value: New LogLevel to use
        """
        ...

    @property
    def verbose(self) -> bool:
        """
        Get current verbosity setting.

        :return: True if verbose mode is enabled.
        """
        ...

    @verbose.setter
    def verbose(self, value: bool) -> None:
        """
        Set the default verbosity setting.

        :param value: New verbosity setting
        """
        ...

    @property
    def precision(self) -> Tuple[int, int]:
        """
        Get current precision level as a negative exponent, i.e. a precision
        of 10^(-3), in real and imaginary parts respectively, corresponds
        to a precision level of (3, 3).

        :return: Current precision level
        """
        ...

    @precision.setter
    def precision(self, value: Tuple[int, int]) -> None:
        """
        Set the default precision level to `value`.

        :param value: New precision level
        """
        ...



"""
Class FinderAlgorithm from the package pyzeal_algorithms.
This module defines a simple root finding algorithm that works on any
continuously differentiable function by constructing a supporting grid in the
complex plane and starting an ordinary Newton algorithm at these support
points. It is expected that this approach underperforms in compared to
algorithms which fully exploit the holomorphic nature of target functions.

Authors:\n
- Philipp Schuette\n
- Luca Wasmuth\n
"""

from itertools import product
from warnings import filterwarnings

import numpy as np
import scipy as sp  # type: ignore

from pyzeal.algorithms.finder_algorithm import FinderAlgorithm
from pyzeal.pyzeal_types.root_types import tVec
from pyzeal.utils.root_context import RootContext


class NewtonGridAlgorithm(FinderAlgorithm):
    """
    Class representation of a root finding algorithm for holomorphic functions
    based on starting an ordinary Newton algorithm on a grid of support points
    in the complex plane.
    """

    __slots__ = ("numSamplePoints",)

    def __init__(self, numSamplePoints: int = 50) -> None:
        r"""
        Initialize a root finding algorithm which searches for roots using the
        Newton algorithm with starting points on an evenly spaced grid.

        :param numSamplePoints: Number of support points in grid rows/columns.
        """
        self.numSamplePoints = numSamplePoints
        self.logger.debug("initialized a new NewtonGridAlgorithm!")

    def calcRoots(self, context: RootContext) -> None:
        """
        Calculate roots in a given context based on the Newton algorithm on a
        grid of support points in the complex plane.

        :param context: Context in which the algorithm operates.
        """
        self.logger.info(
            "starting newton grid search for %s",
            context.functionDataToString(),
        )
        rePoints = np.linspace(
            context.reRan[0],
            context.reRan[1],
            self.numSamplePoints,
            dtype=np.complex128,
        )
        imPoints = np.linspace(
            context.imRan[0],
            context.imRan[1],
            self.numSamplePoints,
            dtype=np.complex128,
        )
        points = [x + y * 1j for (x, y) in product(rePoints, imPoints)]
        filterwarnings("ignore", ".*some failed to converge")
        try:
            roots: tVec = sp.optimize.newton(context.f, points, context.df)
        except RuntimeError:
            return
        for root in roots:
            # the newton algorithm does not determine root orders - placeholder
            # value can be anything non-positive
            context.container.addRoot((root, 0), context.toFilterContext())
        if context.progress and context.task:
            context.progress.update(
                context.task,
                advance=(
                    (context.reRan[1] - context.reRan[0])
                    * (context.imRan[1] - context.imRan[0])
                ),
            )



"""
The `estimators` package contains the various supported estimator modes.

An `ArgumentEstimator` calculates the logarithmic derivative of a holomorphic
function along the boundary of a given rectangular region in the complex plane.
"""

from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.estimators.estimator_cache import EstimatorCache

__all__ = [
    "ArgumentEstimator",
    "EstimatorCache",
]



"""
Various utilities and framework elements used to support `PyZEAL`.
"""

from pyzeal.utils.configuration_exception import InvalidServiceConfiguration
from pyzeal.utils.lambda_wrapper import LambdaWrapper
from pyzeal.utils.service_locator import ServiceLocator

__all__ = [
    "InvalidServiceConfiguration",
    "LambdaWrapper",
    "ServiceLocator",
]



"""
Module constants.py from the package PyZEAL.
This module collects constants used in various root finding algorithms for
easy of reference and consistency.

Authors:\n
- Philipp Schuette\n
"""

from typing import Final

from numpy import pi

# numerical approximation of the discrete boundary between 0 and 2*pi
TWO_PI: Final[float] = 0.8 * (2 * pi)
# numerical approximation of the discrete boundary between 2*pi and 4*pi
FOUR_PI: Final[float] = 0.65 * (4 * pi)

# default values for argument estimation via phase summation
DEFAULT_NUM_PTS = 6500
DEFAULT_DELTA_PHI = 0.01
DEFAULT_MAX_PRECISION = 1e-10

# cutoff for polynomial construction
MAXIMUM_PHASE_MULTIPLIER: Final[int] = 2



"""
Exception class for invalid setting configurations

Authors:\n
- Philipp Schuette\n
"""


class InvalidSettingException(Exception):
    "Raise (and expect) this except whenever an invalid option is encountered."

    def __init__(self, setting: str) -> None:
        """
        Initialize an `InvalidSettingException` instance.

        :param setting: the setting which caused the exception
        """
        super().__init__(f"setting invalid value for {setting}!")



"""
Module installation_helper.py from the package PyZEAL.
This module contains various helper functions related to plugin installation.

Authors:\n
- Philipp Schuette\n
"""

from os import remove
from os.path import exists, join
from shutil import copy

from pyzeal.plugins.plugin_loader import PLUGIN_INSTALL_DIR


class InstallationHelper:
    "Static helper class that provides plugin installation related functions."

    @staticmethod
    def installPlugin(fullPath: str) -> bool:
        """
        Copies the plugin located at `fullPath` to the plugin directory to
        install it.

        :param fullPath: Current path of the to-be-installed plugin.
        :return: `True` if successful.
        """
        if not exists(fullPath):
            return False
        copy(fullPath, PLUGIN_INSTALL_DIR)
        return True

    @staticmethod
    def uninstallPlugin(filename: str) -> bool:
        """
        Uninstall the plugin given by `filename`.

        :param filename: Filename of plugin to remove
        :return: `True` if successful.
        """
        try:
            remove(join(PLUGIN_INSTALL_DIR, filename))
            return True
        except (FileNotFoundError, OSError):
            return False

    @staticmethod
    def returnDataPath(
        filename: str, fullPath: str = PLUGIN_INSTALL_DIR
    ) -> str:
        """
        Helper function to construct the full path to plugin data.

        :param filename: Filename of plugin data.
        :param fullPath: Full path to the directory in which the data is
            located, defaults to PLUGIN_INSTALL_DIR
        :return: Path to plugin data
        """
        return join(fullPath, filename)



"""
This module contains tests of the SIMPLE_ARGUMENT implementation
of the root finding algorithm interface.

Authors:\n
- Luca Wasmuth\n
"""

import numpy as np
import pytest

from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.tests.resources.finder_helpers import simpleArgumentRootFinder
from pyzeal.tests.resources.finder_test_cases import testFunctions
from pyzeal.tests.resources.utils import rootsMatchClosely
from pyzeal.utils.service_locator import ServiceLocator

# disable progress bar by default for tests
settingsService = RAMSettingsService(verbose=False)
ServiceLocator.registerAsSingleton(SettingsService, settingsService)

# some test functions do not work due to z-refinement limitations
KNOWN_FAILURES = ["poly", "x^100", "1e6 * x^100"]


@pytest.mark.parametrize("testName", sorted(testFunctions.keys()))
@pytest.mark.parametrize("parallel", [False, True])
@pytest.mark.parametrize(
    "estimator",
    [EstimatorTypes.SUMMATION_ESTIMATOR, EstimatorTypes.QUADRATURE_ESTIMATOR],
)
def testSimpleArgument(
    testName: str, parallel: bool, estimator: EstimatorTypes
) -> None:
    """
    Test the SIMPLE_ARGUMENT RootFinder with the test case given by
    `testName`.

    :param testName: Name of the test case
    :param parallel: If roots should be searched in parallel
    :param estimator: The type of estimator to use
    """
    if testName in KNOWN_FAILURES:
        pytest.skip()

    reRan = testFunctions[testName].reRan
    imRan = testFunctions[testName].imRan
    if testName == "x^5-4x+2":
        reRan = (-5, 5)
        imRan = (-5, 5)

    hrf = simpleArgumentRootFinder(
        testName, parallel=parallel, estimatorType=estimator
    )
    precision = testFunctions[testName].precision
    hrf.calculateRoots(reRan, imRan, precision=precision)
    foundRoots = hrf.roots
    expectedRoots = np.array(testFunctions[testName].expectedRoots)

    assert rootsMatchClosely(foundRoots, expectedRoots, precision=precision)



"""
Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol

from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_logging.log_manager import LogManager
from pyzeal.pyzeal_logging.logger_facade import PyZEALLogger
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.service_locator import ServiceLocator


class Loggable(Protocol):
    """
    Mixin for combination with classes which support logging, in
    particular changing the logging level.
    """

    __slots__ = ("_logger",)

    _logger: PyZEALLogger

    @property
    def logger(self) -> PyZEALLogger:
        """
        The logger instance associated with this Loggable class. Instance
        creation happens upon first property access.

        :returns: the logger of this class
        """
        if not hasattr(self, "_logger"):
            self._logger = LogManager.initLogger(
                self.__module__.rsplit(".", maxsplit=1)[-1],
                ServiceLocator.tryResolve(SettingsService).logLevel,
            )
        return self._logger

    def setLevel(self, level: LogLevel) -> None:
        """
        Set the log level.

        :param level: the new log level
        """
        self.logger.setLevel(level=level.value)



"""
Module lambda_wrapper.py from the package PyZEAL.
This module wraps lambda functions to allow for multiprocessing, as normally
lambda functions can not be pickled.

Authors:\n
- Philipp Schuette\n
"""

from typing import Literal, Optional, Union

from pyzeal.pyzeal_types.root_types import tHoloFunc, tVec

# global variables holding the lambda function to be wrapped and its derivative
_func: Optional[tHoloFunc] = None
_derivative: Optional[tHoloFunc] = None


def wrappedFunc(z: tVec) -> tVec:
    "The wrapped input function."
    if _func is not None:
        return _func(z)
    raise ValueError("something went wrong during lambda wrapping!")


def wrappedDerivative(z: tVec) -> tVec:
    "The wrapped derivative function."
    if _derivative is not None:
        return _derivative(z)
    raise ValueError("something went wrong during derivative wrapping!")


class LambdaWrapper:
    """
    Static wrapper class for lambda functions.
    """

    @staticmethod
    def wrapLambda(
        func: tHoloFunc,
        mode: Union[Literal["save"], Literal["unsave"]] = "save",
    ) -> tHoloFunc:
        """
        Wrap a lambda function by passing it to `func`.

        :param func: Function to wrap
        :param mode: Wrapping mode, defaults to "save". "unsave" mode will
            overwrite any pre-existing functions, while "save" mode will
            raise an error if one attempts to overwrite the wrapped function.
        :raises ValueError: Raises an error if one tries to wrap a function
            in "save" mode, while another function is already being wrapped.
        :return: The wrapped function.
        """
        global _func

        if _func is not None and mode == "save":
            raise ValueError(
                "you cannot wrap more than one lambda in save mode!"
            )

        _func = func
        return wrappedFunc

    @staticmethod
    def wrapDerivative(
        derivative: tHoloFunc,
        mode: Union[Literal["save"], Literal["unsave"]] = "save",
    ) -> tHoloFunc:
        """
        Wrap a lambda derivative function by passing it to `func`.

        :param func: Derivative function to wrap
        :param mode: Wrapping mode, defaults to "save". "unsave" mode will
            overwrite any pre-existing functions, while "save" mode will
            raise an error if one attempts to overwrite the wrapped function.
        :raises ValueError: Raises an error if one tries to wrap a function
            in "save" mode, while another function is already being wrapped.
        :return: The wrapped derivative.
        """
        global _derivative

        if _derivative is not None and mode == "save":
            raise ValueError(
                "you cannot wrap more than one derivative in save mode!"
            )

        _derivative = derivative
        return wrappedDerivative



"""
Module exporting the public api of the `PyZEAL` project.

The public (scripting) interface to `PyZEAL` comprises the abstract root
finding interface, the concrete root finders, and the types (enums) necessary
to instantiate concrete root finders.
"""

from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.rootfinders.finder_interface import RootFinderInterface
from pyzeal.rootfinders.parallel_finder import ParallelRootFinder
from pyzeal.rootfinders.rootfinder import RootFinder

__all__ = [
    "RootFinderInterface",
    "RootFinder",
    "ParallelRootFinder",
    "AlgorithmTypes",
    "ContainerTypes",
    "EstimatorTypes",
]



"""
This module provides named constants used to identify concrete argument
estimator implementations in any instance where such an estimator is required.

Authors:\n
- Philipp Schuette\n
"""

from enum import Enum


class EstimatorTypes(Enum):
    "Enumeration containing named constants identifying available estimators."
    SUMMATION_ESTIMATOR = "SummationEstimator"
    QUADRATURE_ESTIMATOR = "QuadratureEstimator"
    DEFAULT = "DefaultEstimator"



"""
This module contains tests for the plain container implementation.
"""

from typing import List

import numpy as np

from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.filter_context import FilterContext


def testAddPlainContainer(
    roundingContainer: RootContainer, filterContexts: List[FilterContext]
) -> None:
    """
    Test the plain container `addRoot` implementation.
    """
    assert len(roundingContainer.getRoots()) == 0
    # filter contexts differ in precision, effectively clearing container
    for filterContext in filterContexts:
        roundingContainer.addRoot((1.23456789, 1), filterContext)
        assert len(roundingContainer.getRoots()) == 1

        # root should be rejected due to equality
        roundingContainer.addRoot((1.2345678, 1), filterContext)
        assert len(roundingContainer.getRoots()) == 1

        roundingContainer.addRoot((-1.23456789, 2), filterContext)
        assert len(roundingContainer.getRoots()) == 2

        # root should not be rejected due to inequality of orders
        roundingContainer.addRoot((1.2345678, 2), filterContext)
        assert len(roundingContainer.getRoots()) == 3

    containerContents = np.sort(roundingContainer.getRoots())
    assert (containerContents == [-1.235, 1.235, 1.235]).all()


def testRemovePlainContainer(
    roundingContainer: RootContainer, filterContexts: List[FilterContext]
) -> None:
    """
    Test the plain container `removeRoot` implementation.
    """
    roundingContainer.addRoot((0, 1), filterContexts[0])
    roundingContainer.addRoot((1, 1), filterContexts[0])

    assert len(roundingContainer.getRoots()) == 2
    assert not roundingContainer.removeRoot((-1, 1))
    assert roundingContainer.removeRoot((0, 1))
    assert len(roundingContainer.getRoots()) == 1
    assert roundingContainer.removeRoot((1, 1))
    assert len(roundingContainer.getRoots()) == 0
    assert not roundingContainer.removeRoot((0, 1))


def testClearPlainContainer(
    roundingContainer: RootContainer, filterContexts: List[FilterContext]
) -> None:
    """
    Test the plain container `clear` implementation.
    """
    roundingContainer.addRoot((0, 1), filterContexts[0])
    roundingContainer.addRoot((-2, 1), filterContexts[0])

    assert len(roundingContainer.getRoots()) == 2
    roundingContainer.clear()
    assert len(roundingContainer.getRoots()) == 0


def testRegisterUnregisterFilterPlainContainer(
    roundingContainer: RootContainer, filterContexts: List[FilterContext]
) -> None:
    """
    Test the plain container `registerFilter` implementation.
    """
    roundingContainer.registerFilter(lambda r, c: False, "alwaysFalse")
    roundingContainer.addRoot((1j, 0), filterContexts[0])
    assert len(roundingContainer.getRoots()) == 0

    roundingContainer.unregisterFilter("alwaysFalse")
    roundingContainer.addRoot((1j, 0), filterContexts[0])
    assert len(roundingContainer.getRoots()) == 1



"""
Class RootContext from the package pyzeal_util.
This module defines a data container that holds the information necessary for
a generic root finding algorithm to function.

Authors:\n
- Philipp Schuette\n
"""

from dataclasses import dataclass
from typing import Optional, Tuple

from rich.progress import TaskID

from pyzeal.pyzeal_types.root_types import tHoloFunc
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.filter_context import FilterContext
from pyzeal.utils.finder_progress import FinderProgressBar


@dataclass(frozen=True)
class RootContext:
    """
    Container for the data context of a root finding algorithm. The container
    is read-only.
    """

    f: tHoloFunc
    df: Optional[tHoloFunc]
    container: RootContainer
    precision: Tuple[int, int]
    reRan: Tuple[float, float] = (-1.0, 1.0)
    imRan: Tuple[float, float] = (-1.0, 1.0)
    progress: Optional[FinderProgressBar] = None
    task: Optional[TaskID] = None

    def toFilterContext(self) -> FilterContext:
        """
        Get a `FilterContext` object with the same parameters as this
        `RootContext`

        :return: `FilterContext` object with the same parameters as this
            `RootContext`
        """
        return FilterContext(self.f, self.reRan, self.imRan, self.precision)

    def functionDataToString(self) -> str:
        """
        Return a string describing the data stored by this object

        :return: Object data
        """
        return (
            f"{getattr(self.f, '__name__', '<unnamed>')} on rectangle "
            + f"[{self.reRan[0]}, {self.reRan[1]}] "
            + f"x [{self.imRan[0]}, {self.imRan[1]}]"
        )



"""
Protocol RootContainer from the package pyzeal_utils.
This module defines an interface to container types which are meant to hold
root data. The underlying strategy used in cases where e.g. almost identical
roots are added or the precision of roots is changed between add operations is
up to the concrete container implementation (and transparent to the user).
Additional checks could also be implemented in subclasses, e.g. checking if a
root is contained in a certain region.

Authors:\n
- Philipp Schuette\n
"""

from typing import Protocol, runtime_checkable

import numpy as np
from numpy.typing import NDArray

from pyzeal.pyzeal_logging.loggable import Loggable
from pyzeal.pyzeal_types.root_types import tRoot, tVec
from pyzeal.utils.filter_context import FilterContext, tRootFilter


@runtime_checkable
class RootContainer(Loggable, Protocol):
    "Structural interface for container classes meant to hold root data."

    def addRoot(self, root: tRoot, context: FilterContext) -> None:
        """
        Add a root to the container.

        :param root: the root to be added to the container
        :param context: the number of valid decimal places of `root`
        """
        ...

    def getRoots(self) -> tVec:
        """
        Returns all roots currently held in this container as a vector.

        :return: a vector of complex roots
        """
        ...

    def getRootOrders(self) -> NDArray[np.int32]:
        """
        Returns the orders of all roots currently held in this container as a
        vector which is parallel to the vector returned by `getRoots`.

        :return: a vector of integer root orders (multiplicities)
        """
        ...

    def removeRoot(self, root: tRoot) -> bool:
        """
        Remove a root from the container and indicate whether or not a removal
        actually happened via the return value.

        :param root: the root to be removed from the container
        :return: a boolean flag indicating if a removal happened
        """
        ...

    def clear(self) -> None:
        "Clear all data from the container."
        ...

    def registerFilter(self, filterPredicate: tRootFilter, key: str) -> None:
        """
        Add a new filter predicate to this container instance. The filtering
        applies to all roots added after the filter is registered. To guarantee
        a consistent set of roots any implementation should clear its internal
        root buffer before/after registering a filter.

        :param filterPredicate: the predicate used to filter roots
        :param key: a (unique) key under which the filter is registered
        """
        ...

    def unregisterFilter(self, key: str) -> None:
        """
        Remove a previously registered filter by key.

        :param key: key value for the filter to remove
        """
        ...



"""
Class FinderAlgorithm from the package pyzeal_algorithms.

This module defines an abstract class/interface for a generic root finding
algorithm. This builds the foundation for the primary user-facing API of
`PyZEAL`.

Authors:\n
- Philipp Schuette\n
"""

from abc import ABC, abstractmethod

from pyzeal.pyzeal_logging.loggable import Loggable
from pyzeal.utils.root_context import RootContext


class FinderAlgorithm(ABC, Loggable):
    """
    Abstract class representation of a generic root finding algorithm. Concrete
    algorithms are implemented by subclassing this class and overriding the
    virtual method `calcRoots` appropriately.
    """

    @abstractmethod
    def calcRoots(self, context: RootContext) -> None:
        """
        Entry point for a generic root finding algorithm operating in a given
        context. Found roots are expected to be inserted into
        `context.container` upon the algorithms completion.

        :param context: Context in which the algorithm operates.
        """



"""
TODO.

Authors:\n
- Philipp Schuette\n
"""

from os.path import abspath
from typing import Optional, Tuple

from pyzeal.cli.controller_facade import CLIControllerFacade
from pyzeal.cli.parse_results import (
    InstallTestingParseResults,
    PluginParseResults,
    SettingsParseResults,
)
from pyzeal.plugins.installation_helper import InstallationHelper
from pyzeal.plugins.plugin_loader import PluginLoader
from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.install_test_facade import InstallTestingHandlerFacade


class CLIController(CLIControllerFacade):
    """
    Controller class handling the business logic of the `PyZEAL` command line
    interface.
    """

    def __init__(
        self,
        testingHandler: InstallTestingHandlerFacade,
        settings: SettingsService,
    ) -> None:
        """
        Initialize a new command line interface controller instance.
        """
        self.settingsService = settings
        self.testingHandler = testingHandler

    # docstr-coverage:inherited
    def handleViewSubcommand(self, args: SettingsParseResults) -> None:
        if args.doPrint:
            print(self.settingsService)

    # docstr-coverage:inherited
    def handleChangeSubcommand(self, args: SettingsParseResults) -> None:
        settingsService = self.settingsService
        if args.container:
            self.changeContainerSetting(
                args.container + "_container", settingsService
            )
        if args.algorithm:
            self.changeAlgorithmSetting(args.algorithm, settingsService)
        if args.estimator:
            self.changeEstimatorSetting(
                args.estimator + "_estimator", settingsService
            )
        if args.logLevel:
            self.changeLogLevelSetting(args.logLevel, settingsService)
        if args.verbose:
            self.changeVerbositySetting(args.verbose, settingsService)
        if args.precision:
            self.changePrecisionSetting(args.precision, settingsService)

    # docstr-coverage:inherited
    def handlePluginSubcommand(self, args: PluginParseResults) -> None:
        if args.listPlugins:
            print("currently installed Plugins:")
            print("----------------------------")
            for plugin in PluginLoader.loadPlugins():
                print(plugin)

        if args.listModules:
            print("installation directory contents:")
            print("--------------------------------")
            for module in PluginLoader.discoverModules():
                print(module)

        if args.install:
            print(f"installing plugin {args.install}...")
            if InstallationHelper.installPlugin(abspath(args.install)):
                print("[success] plugin was installed!")
            else:
                print("[error] the requested plugin does not exist - abort!")
                raise SystemExit(2)

        if args.uninstall:
            print(f"uninstalling plugin {args.uninstall}...")
            if InstallationHelper.uninstallPlugin(args.uninstall):
                print("[success] plugin was uninstalled!")
            else:
                print("[error] the requested plugin does not exist - abort!")
                raise SystemExit(2)

    # docstr-coverage:inherited
    def handleTestingOption(self, args: InstallTestingParseResults) -> bool:
        if args.doTest:
            # right now we do not include root finder tests here to reduce
            # runtime; one could include those tests after performance increase
            modules = ["cli", "algorithms", "containers", "estimators"]
            for module in modules:
                self.testingHandler.testModule(module=module)
            return True
        return False

    @staticmethod
    def changeContainerSetting(
        container: str, service: SettingsService
    ) -> None:
        """
        Try to change the default container setting in `service` to
        `container`. If the container name is invalid, `SystemExit(2)` is
        raised.

        :param container: New default container name (case-insensitive).
        :param service: Settings service to update.
        :raises SystemExit: Raised when no matching container can be found.
        """
        oldContainer = service.defaultContainer
        newContainer: Optional[ContainerTypes] = None
        for containerType in ContainerTypes:
            if containerType.name == container.upper():
                newContainer = containerType
                break
        if newContainer is None:
            raise SystemExit(2)
        if newContainer != oldContainer:
            service.defaultContainer = newContainer
            print(
                "changed default container:   "
                + oldContainer.value
                + " --> "
                + oldContainer.value
            )

    @staticmethod
    def changeAlgorithmSetting(
        algorithm: str, service: SettingsService
    ) -> None:
        """
        Try to change the default algorithm setting in `service` to
        `algorithm`. If the algorithm name is invalid, `SystemExit(2)` is
        raised.

        :param algorithm: New default algorithm name (case-insensitive)
        :param service: Settings service to update.
        :raises SystemExit: Raised when no matching algorithm can be found.
        """
        oldAlgorithm = service.defaultAlgorithm
        newAlgorithm: Optional[AlgorithmTypes] = None
        for algorithmType in AlgorithmTypes:
            if algorithmType.name == algorithm.upper():
                newAlgorithm = algorithmType
                break
        if newAlgorithm is None:
            raise SystemExit(2)
        if newAlgorithm != oldAlgorithm:
            service.defaultAlgorithm = newAlgorithm
            print(
                "changed default algorithm:   "
                + oldAlgorithm.value
                + " --> "
                + newAlgorithm.value
            )

    @staticmethod
    def changeEstimatorSetting(
        estimator: str, service: SettingsService
    ) -> None:
        """
        Try to change the default estimator setting in `service` to
        `estimator`. If the estimator name is invalid, `SystemExit(2)` is
        raised.

        :param estimator: New default estimator name (case-insensitive)
        :param service: Settings service to update.
        :raises SystemExit: Raised when no matching algorithm can be found.
        """
        oldEstimator = service.defaultEstimator
        newEstimator: Optional[EstimatorTypes] = None
        for estimatorType in EstimatorTypes:
            if estimatorType.name == estimator.upper():
                newEstimator = estimatorType
                break
        if newEstimator is None:
            raise SystemExit(2)
        if newEstimator != oldEstimator:
            service.defaultEstimator = newEstimator
            print(
                "changed default estimator:   "
                + oldEstimator.value
                + " --> "
                + newEstimator.value
            )

    @staticmethod
    def changeLogLevelSetting(logLevel: str, service: SettingsService) -> None:
        """
        Try to change the logLevel setting in `service` to `logLevel`.
        If the logLevel is invalid, `SystemExit(2)` is raised.

        :param logLevel: New log level (case-insensitive)
        :param service: Settings service to update.
        :raises SystemExit: Raised when the log level is invalid.
        """
        oldLevel = service.logLevel
        newLevel: Optional[LogLevel] = None
        for level in LogLevel:
            if level.name == logLevel.upper():
                newLevel = level
                break
        if newLevel is None:
            raise SystemExit(2)
        if newLevel != oldLevel:
            service.logLevel = newLevel
            print(
                "changed default log level:   "
                + oldLevel.name
                + " --> "
                + newLevel.name
            )

    @staticmethod
    def changeVerbositySetting(verbose: str, service: SettingsService) -> None:
        """
        Try to update the verbosity setting of `service`. If `verbose` is not
        "true" or "false", `SystemExit(2)` is raised.

        :param verbose: New verbosity setting
        :param service: Setting service to update
        """
        oldVerbosity = service.verbose
        newVerbosity: bool = verbose.lower() == "true"
        if newVerbosity != oldVerbosity:
            service.verbose = newVerbosity
            print(
                "changed default verbosity:   "
                + str(oldVerbosity)
                + " --> "
                + str(newVerbosity)
            )

    @staticmethod
    def changePrecisionSetting(
        precision: Tuple[int, int], service: SettingsService
    ) -> None:
        """
        Try to update the precision setting of `service`. If `precision` is not
        a pair of ints, `SystemExit(2)` is raised.

        :param precision: New precision setting
        :param service: Setting service to update
        """
        oldPrecision = service.precision
        newPrecision: Tuple[int, int] = precision
        if newPrecision != oldPrecision:
            service.precision = newPrecision
            print(
                "changed default precision:   "
                + str(oldPrecision)
                + " --> "
                + str(newPrecision)
            )



"""
This module contains tests for the command-line interface, apart from the
parsing.
"""

from functools import partial
from typing import Callable, Tuple, TypedDict, Union
from unittest.mock import PropertyMock, create_autospec, patch

import pytest

from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.finder_algorithm import FinderAlgorithm
from pyzeal.cli.__main__ import PyZEALEntry
from pyzeal.cli.cli_controller import CLIController
from pyzeal.cli.cli_parser import PyZEALParser
from pyzeal.cli.controller_facade import CLIControllerFacade
from pyzeal.cli.parse_results import (
    InstallTestingParseResults,
    PluginParseResults,
    SettingsParseResults,
)
from pyzeal.cli.parser_facade import PyZEALParserInterface
from pyzeal.pyzeal_logging.log_levels import LogLevel
from pyzeal.pyzeal_types.algorithm_types import AlgorithmTypes
from pyzeal.pyzeal_types.container_types import ContainerTypes
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.settings.ram_settings_service import RAMSettingsService
from pyzeal.settings.settings_service import SettingsService
from pyzeal.utils.containers.root_container import RootContainer
from pyzeal.utils.factories.algorithm_factory import AlgorithmFactory
from pyzeal.utils.factories.container_factory import ContainerFactory
from pyzeal.utils.factories.estimator_factory import EstimatorFactory
from pyzeal.utils.install_test_facade import InstallTestingHandlerFacade
from pyzeal.utils.install_test_handler import InstallTestingHandler
from pyzeal.utils.service_locator import ServiceLocator

ServiceLocator.registerAsSingleton(
    PyZEALParserInterface, PyZEALParser()
).registerAsTransient(
    FinderAlgorithm, AlgorithmFactory.getConcreteAlgorithm
).registerAsTransient(
    RootContainer, ContainerFactory.getConcreteContainer
).registerAsTransient(
    ArgumentEstimator, EstimatorFactory.getConcreteEstimator
).registerAsTransient(
    CLIControllerFacade, CLIController
).registerAsTransient(
    InstallTestingHandlerFacade, InstallTestingHandler
).registerAsSingleton(
    SettingsService, RAMSettingsService()
)


class SettingsDict(TypedDict, total=False):
    """
    Container class to ensure correct typing of settings options to be tested.
    """

    container: str
    algorithm: str
    logLevel: str
    verbose: str
    estimator: str
    precision: Tuple[int, int]


def mockArgs(
    newSetting: SettingsDict,
) -> Tuple[
    SettingsParseResults, PluginParseResults, InstallTestingParseResults
]:
    """
    Generate `SettingsParseResults` and `PluginParseResults` instances
    containing the settings given by `newSetting`.

    :param newSetting: Dict of settings and values
    :return: ParseResults with appropriate settings
    """
    settingsParseResult = SettingsParseResults(
        doPrint=True,
        container=newSetting.get("container", ""),
        algorithm=newSetting.get("algorithm", ""),
        logLevel=newSetting.get("logLevel", ""),
        verbose=newSetting.get("verbose", ""),
        estimator=newSetting.get("estimator", ""),
        precision=newSetting.get("precision", None),
    )
    pluginParseResult = PluginParseResults(
        listPlugins=False,
        listModules=False,
        install="",
        uninstall="",
    )
    testingParseResult = InstallTestingParseResults(doTest=False)
    return settingsParseResult, pluginParseResult, testingParseResult


tSettingsTypes = Union[
    AlgorithmTypes,
    bool,
    ContainerTypes,
    LogLevel,
    EstimatorTypes,
    Tuple[int, int],
]

newSettings: Tuple[SettingsDict, ...] = (
    {"algorithm": "NEWTON_GRID"},
    {"verbose": "True"},
    {"container": "plain"},
    {"logLevel": "INFO"},
    {"estimator": "summation"},
    {"precision": (3, 3)},
)
changeMethods: Tuple[str, ...] = (
    "changeAlgorithmSetting",
    "changeVerbositySetting",
    "changeContainerSetting",
    "changeLogLevelSetting",
    "changeEstimatorSetting",
    "changePrecisionSetting",
)
defaultSettingProperties: Tuple[str, ...] = (
    "defaultAlgorithm",
    "verbose",
    "defaultContainer",
    "logLevel",
    "defaultEstimator",
    "precision",
)
beforeValues: Tuple[tSettingsTypes, ...] = (
    AlgorithmTypes.SIMPLE_ARGUMENT,
    False,
    ContainerTypes.ROUNDING_CONTAINER,
    LogLevel.WARNING,
    EstimatorTypes.QUADRATURE_ESTIMATOR,
    (1, 1),
)
afterValues: Tuple[tSettingsTypes, ...] = (
    AlgorithmTypes.NEWTON_GRID,
    True,
    ContainerTypes.PLAIN_CONTAINER,
    LogLevel.INFO,
    EstimatorTypes.SUMMATION_ESTIMATOR,
    (3, 3),
)


@pytest.mark.parametrize("testSetup", zip(newSettings, changeMethods))
def testChangeCall(testSetup: Tuple[SettingsDict, str]) -> None:
    """
    Test if `change...Setting` gets called correctly.
    """
    newSetting: SettingsDict = testSetup[0]
    changeMethod = "pyzeal.cli.cli_controller.CLIController." + testSetup[1]
    with patch("pyzeal.cli.cli_parser.PyZEALParser.parseArgs") as mockParse:
        mockParse.side_effect = partial(mockArgs, newSetting=newSetting)
        with patch(changeMethod) as doNothing:
            entry = PyZEALEntry()
            entry.mainPyZEAL()
            doNothing.assert_called_once()


@pytest.mark.parametrize(
    "testSetup",
    zip(newSettings, defaultSettingProperties, beforeValues, afterValues),
)
def testChangeSettingsCall(
    testSetup: Tuple[
        SettingsDict,
        str,
        tSettingsTypes,
        tSettingsTypes,
    ]
) -> None:
    """
    Test if `change...Setting` correctly updates the newSetting.
    """
    newSetting: SettingsDict = testSetup[0]
    defaultSettingProperty = (
        "pyzeal.settings.ram_settings_service.RAMSettingsService."
        + testSetup[1]
    )
    beforeValue = testSetup[2]
    afterValue = testSetup[3]
    with patch("pyzeal.cli.cli_parser.PyZEALParser.parseArgs") as mockParse:
        mockParse.side_effect = partial(mockArgs, newSetting=newSetting)
        with patch(
            defaultSettingProperty, new_callable=PropertyMock
        ) as mockedProp:
            mockedProp.return_value = beforeValue
            entry = PyZEALEntry()
            entry.mainPyZEAL()
            mockedProp.assert_called_with(afterValue)


@pytest.mark.parametrize(
    "changeFunction",
    [
        CLIController.changeAlgorithmSetting,
        CLIController.changeContainerSetting,
        CLIController.changeEstimatorSetting,
        CLIController.changeLogLevelSetting,
    ],
)
def testCLIControllerInvalidName(
    changeFunction: Callable[[str, SettingsService], None]
) -> None:
    """
    Test if CLIController correctly exits when an invalid name is given
    for a newSetting.

    :param changeFunction: Function to test, parametrized by pytest
    """
    mockSettings = create_autospec(
        SettingsService, spec_set=True, instance=True
    )
    with pytest.raises(SystemExit):
        changeFunction("THIS_DOES_NOT_EXIST", mockSettings)



"""
Class SimpleHoloAlgorithm from the package pyzeal_algorithms.
This module defines a root finding algorithm based on a simple, straight
forward approach to the argument principle. The simplicity comes from the fact
that we avoid numeric quadrature by approximating integration of the
logarithmic derivative via changes in the complex argument of the target
function. Due to the possibility of overlooking large changes in the complex
argument this algorithm is not completely reliable, see [Henrici].

Authors:\n
- Philipp Schuette\n
"""

from typing import Tuple

import numpy as np

from pyzeal.algorithms.constants import (
    DEFAULT_DELTA_PHI,
    DEFAULT_MAX_PRECISION,
    DEFAULT_NUM_PTS,
    TWO_PI,
)
from pyzeal.algorithms.estimators import EstimatorCache
from pyzeal.algorithms.estimators.argument_estimator import ArgumentEstimator
from pyzeal.algorithms.finder_algorithm import FinderAlgorithm
from pyzeal.pyzeal_types.estimator_types import EstimatorTypes
from pyzeal.utils.root_context import RootContext
from pyzeal.utils.service_locator import ServiceLocator


class SimpleArgumentAlgorithm(FinderAlgorithm):
    """
    Class representation of a simple root finding algorithm for holomorphic
    functions based on the argument principle and the approximation of
    integrals over logarithmic derivatives via the summation of phase
    differences.
    """

    __slots__ = ("cache", "estimator")

    def __init__(
        self,
        estimatorType: EstimatorTypes,
        *,
        numPts: int = DEFAULT_NUM_PTS,
        deltaPhi: float = DEFAULT_DELTA_PHI,
        maxPrecision: float = DEFAULT_MAX_PRECISION,
    ) -> None:
        """
        Initialize a root finding algorithm that employs a straightforward,
        simple adaptation of the argument principle by refining an initial
        bounding rectangle until it either contains no further roots or it is
        smaller in size than the requested accuracy.

        :param numPts: the default number of support points on rectangle edges
            at the start of dynamic refinement
        :param deltaPhi: the maximal phase shift between neighboring points on
            rectangle edges before dynamic refinement starts
        :param maxPrecision: the minimal distance between neighboring points on
            rectangle edges during dynamic refinement
        """
        self.cache = EstimatorCache()
        self.estimator = ServiceLocator.tryResolve(
            ArgumentEstimator,
            estimatorType=estimatorType,
            numPts=numPts,
            deltaPhi=deltaPhi,
            maxPrecision=maxPrecision,
            cache=self.cache,
        )
        self.logger.debug(
            "initialized a new subclass of SimpleArgumentAlgorithm!"
        )

    def calcRoots(self, context: RootContext) -> None:
        """
        Start a root calculation using a straightforward argument principle
        based refinement strategy.

        This routine calculates the initially expected number of roots by
        delegating the argument calculation to its `ArgumentEstimator`
        instance. Then it delegates the actual work of recursive refinement to
        the routines `decideRefinement` and `calcRootsRecursion`.

        :param context: context in which the algorithm operates
        """
        self.logger.info(
            "starting simple argument search for %s",
            context.functionDataToString(),
        )
        # reset cache
        if self.cache.dirty():
            self.logger.info("resetting argument estimator cache...")
            self.cache.reset()

        phi = self.estimator.calcMoment(
            0, context.reRan, context.imRan, context
        ).real  # if order=0 then the result is (theoretically) an int
        self.decideRefinement(context.reRan, context.imRan, phi, context)
        self.logger.debug(
            "cache hits/misses: %d/%d (= %.03f)",
            self.cache.cacheHits,
            self.cache.cacheMisses + self.cache.cacheHits,
            round(
                self.cache.cacheHits
                / (self.cache.cacheMisses + self.cache.cacheMisses or 1),
                3,
            ),
        )

    def decideRefinement(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        phi: float,
        context: RootContext,
    ) -> None:
        """
        Decide which way the current search area should be subdivided and
        calculate roots in the subdivided areas. The simple strategy consists
        of the choices (1) return, if argument indicates no roots, (2) place
        root in `context.container` if roots present and accuracy attained, or
        (3) subdivide further if roots present but accuracy not yet attained.

        :param reRan: Real part of current search Range
        :param imRan: Imaginary part of current search range
        :param phi: Change in argument along the boundary of the current range
        :param context: `RootContext` in which the algorithm operates
        """
        # calculate difference between right/left and top/bottom
        x1, x2 = reRan
        y1, y2 = imRan
        deltaRe = x2 - x1
        deltaIm = y2 - y1

        # check if current rectangle contains zeros
        if phi < TWO_PI:
            if context.progress is not None and context.task is not None:
                context.progress.update(
                    context.task, advance=deltaRe * deltaIm
                )
            return

        # the current rectangle does contain (at least one) zero(s)
        self.logger.debug(
            "Rectangle [%s, %s] x [%s, %s] contains zeros!",
            str(x1),
            str(x2),
            str(y1),
            str(y2),
        )
        self.logger.debug("Rectangle diameters are %f x %f", deltaRe, deltaIm)

        # check if desired accuracy is aquired
        epsReal = 10 ** (-context.precision[0])
        epsImag = 10 ** (-context.precision[1])
        if deltaRe < epsReal and deltaIm < epsImag:
            SimpleArgumentAlgorithm.getRootFromRectangle(
                x2, x1, y2, y1, phi, context
            )
            return

        # the current box contains a root and must be refined further
        if deltaRe / epsReal > deltaIm / epsImag:
            midPoint = (x1 + x2) / 2
            phi = self.calculateRefinedMoment(
                (x1, midPoint), (y1, y2), context
            )
            self.decideRefinement((x1, midPoint), (y1, y2), phi, context)
            self.cache.remove(0, x1 + y2 * 1j, x1 + y1 * 1j)
            phi = self.calculateRefinedMoment(
                (midPoint, x2), (y1, y2), context
            )
            self.decideRefinement((midPoint, x2), (y1, y2), phi, context)
            self.cache.remove(0, x2 + y1 * 1j, x2 + y2 * 1j)
        else:
            midPoint = (y1 + y2) / 2
            phi = self.calculateRefinedMoment(
                (x1, x2), (y1, midPoint), context
            )
            self.decideRefinement((x1, x2), (y1, midPoint), phi, context)
            self.cache.remove(0, x1 + y1 * 1j, x2 + y1 * 1j)
            phi = self.calculateRefinedMoment(
                (x1, x2), (midPoint, y2), context
            )
            self.decideRefinement((x1, x2), (midPoint, y2), phi, context)
            self.cache.remove(0, x2 + y2 * 1j, x1 + y2 * 1j)

    def calculateRefinedMoment(
        self,
        reRan: Tuple[float, float],
        imRan: Tuple[float, float],
        context: RootContext,
    ) -> float:
        """
        Recursively calculate roots in the given search range according to
        `context`.

        :param reRan: Real part of current search range
        :param imRan: Imaginary part of current search range
        :param context: `RootContext` in which the algorithm operates
        """
        # check if the given rectangle contains at least one zero
        phi: float = self.estimator.calcMoment(
            0, reRan=reRan, imRan=imRan, context=context
        ).real
        self.logger.debug(
            "simple argument recursively refined with total phase=%f",
            phi / (2.0 * np.pi),
        )

        return phi

    @staticmethod
    def getRootFromRectangle(
        right: float,
        left: float,
        top: float,
        bottom: float,
        phi: float,
        context: RootContext,
    ) -> None:
        """
        Compute the location of a root in a sufficiently small rectangle and
        update the progress bar accordingly.

        :param right: right boundary of the rectangle
        :param left: left boundary of the rectangle
        :param top: top boundary of the rectangle
        :param bottom: bottom boundary of the rectangle
        :param phi: the total argument within the rectangle
        :param context: overall context of the current calculation
        """
        newZero = (left + right + 1j * (bottom + top)) / 2.0
        zOrder = int(np.round(phi / (2 * np.pi)))
        context.container.addRoot((newZero, zOrder), context.toFilterContext())
        if context.progress is not None and context.task is not None:
            context.progress.update(
                context.task,
                advance=(right.real - left.real) * (top.imag - bottom.imag),
            )
